from locan.data import metadata_pb2 as _metadata_pb2
from google.protobuf.internal import containers as _containers
from google.protobuf.internal import enum_type_wrapper as _enum_type_wrapper
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from typing import (
    ClassVar as _ClassVar,
    Iterable as _Iterable,
    Mapping as _Mapping,
    Optional as _Optional,
    Union as _Union,
)

DESCRIPTOR: _descriptor.FileDescriptor

class Type(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    TABLE: _ClassVar[Type]
    IMAGE: _ClassVar[Type]

class Mode(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    BINARY: _ClassVar[Mode]
    TEXT: _ClassVar[Mode]

class Dtype(int, metaclass=_enum_type_wrapper.EnumTypeWrapper):
    __slots__ = ()
    INT8: _ClassVar[Dtype]
    UINT8: _ClassVar[Dtype]
    INT16: _ClassVar[Dtype]
    UINT16: _ClassVar[Dtype]
    INT32: _ClassVar[Dtype]
    UINT32: _ClassVar[Dtype]
    INT64: _ClassVar[Dtype]
    UINT64: _ClassVar[Dtype]
    FLOAT32: _ClassVar[Dtype]
    FLOAT64: _ClassVar[Dtype]

TABLE: Type
IMAGE: Type
BINARY: Mode
TEXT: Mode
INT8: Dtype
UINT8: Dtype
INT16: Dtype
UINT16: Dtype
INT32: Dtype
UINT32: Dtype
INT64: Dtype
UINT64: Dtype
FLOAT32: Dtype
FLOAT64: Dtype

class Format(_message.Message):
    __slots__ = (
        "name",
        "type",
        "mode",
        "columns",
        "headers",
        "dtype",
        "shape",
        "units",
        "description",
    )
    NAME_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    MODE_FIELD_NUMBER: _ClassVar[int]
    COLUMNS_FIELD_NUMBER: _ClassVar[int]
    HEADERS_FIELD_NUMBER: _ClassVar[int]
    DTYPE_FIELD_NUMBER: _ClassVar[int]
    SHAPE_FIELD_NUMBER: _ClassVar[int]
    UNITS_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    name: str
    type: Type
    mode: Mode
    columns: int
    headers: _containers.RepeatedScalarFieldContainer[str]
    dtype: _containers.RepeatedScalarFieldContainer[Dtype]
    shape: _containers.RepeatedScalarFieldContainer[int]
    units: _containers.RepeatedScalarFieldContainer[str]
    description: str
    def __init__(
        self,
        name: _Optional[str] = ...,
        type: _Optional[_Union[Type, str]] = ...,
        mode: _Optional[_Union[Mode, str]] = ...,
        columns: _Optional[int] = ...,
        headers: _Optional[_Iterable[str]] = ...,
        dtype: _Optional[_Iterable[_Union[Dtype, str]]] = ...,
        shape: _Optional[_Iterable[int]] = ...,
        units: _Optional[_Iterable[str]] = ...,
        description: _Optional[str] = ...,
    ) -> None: ...

class FileInfo(_message.Message):
    __slots__ = (
        "name",
        "type",
        "format",
        "channel",
        "rows",
        "offset",
        "min",
        "max",
        "exposure",
    )

    class OffsetEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: int
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[int] = ...
        ) -> None: ...

    class MinEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: float
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[float] = ...
        ) -> None: ...

    class MaxEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: float
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[float] = ...
        ) -> None: ...

    NAME_FIELD_NUMBER: _ClassVar[int]
    TYPE_FIELD_NUMBER: _ClassVar[int]
    FORMAT_FIELD_NUMBER: _ClassVar[int]
    CHANNEL_FIELD_NUMBER: _ClassVar[int]
    ROWS_FIELD_NUMBER: _ClassVar[int]
    OFFSET_FIELD_NUMBER: _ClassVar[int]
    MIN_FIELD_NUMBER: _ClassVar[int]
    MAX_FIELD_NUMBER: _ClassVar[int]
    EXPOSURE_FIELD_NUMBER: _ClassVar[int]
    name: str
    type: Type
    format: str
    channel: str
    rows: int
    offset: _containers.ScalarMap[str, int]
    min: _containers.ScalarMap[str, float]
    max: _containers.ScalarMap[str, float]
    exposure: float
    def __init__(
        self,
        name: _Optional[str] = ...,
        type: _Optional[_Union[Type, str]] = ...,
        format: _Optional[str] = ...,
        channel: _Optional[str] = ...,
        rows: _Optional[int] = ...,
        offset: _Optional[_Mapping[str, int]] = ...,
        min: _Optional[_Mapping[str, float]] = ...,
        max: _Optional[_Mapping[str, float]] = ...,
        exposure: _Optional[float] = ...,
    ) -> None: ...

class Manifest(_message.Message):
    __slots__ = (
        "format_version",
        "formats",
        "files",
        "name",
        "description",
        "tags",
        "thumbnail",
        "sample",
        "labeling",
        "date",
        "author",
        "citation",
        "email",
        "map",
        "locdata_meta",
    )

    class FormatsEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: Format
        def __init__(
            self,
            key: _Optional[str] = ...,
            value: _Optional[_Union[Format, _Mapping]] = ...,
        ) -> None: ...

    class MapEntry(_message.Message):
        __slots__ = ("key", "value")
        KEY_FIELD_NUMBER: _ClassVar[int]
        VALUE_FIELD_NUMBER: _ClassVar[int]
        key: str
        value: str
        def __init__(
            self, key: _Optional[str] = ..., value: _Optional[str] = ...
        ) -> None: ...

    FORMAT_VERSION_FIELD_NUMBER: _ClassVar[int]
    FORMATS_FIELD_NUMBER: _ClassVar[int]
    FILES_FIELD_NUMBER: _ClassVar[int]
    NAME_FIELD_NUMBER: _ClassVar[int]
    DESCRIPTION_FIELD_NUMBER: _ClassVar[int]
    TAGS_FIELD_NUMBER: _ClassVar[int]
    THUMBNAIL_FIELD_NUMBER: _ClassVar[int]
    SAMPLE_FIELD_NUMBER: _ClassVar[int]
    LABELING_FIELD_NUMBER: _ClassVar[int]
    DATE_FIELD_NUMBER: _ClassVar[int]
    AUTHOR_FIELD_NUMBER: _ClassVar[int]
    CITATION_FIELD_NUMBER: _ClassVar[int]
    EMAIL_FIELD_NUMBER: _ClassVar[int]
    MAP_FIELD_NUMBER: _ClassVar[int]
    LOCDATA_META_FIELD_NUMBER: _ClassVar[int]
    format_version: str
    formats: _containers.MessageMap[str, Format]
    files: _containers.RepeatedCompositeFieldContainer[FileInfo]
    name: str
    description: str
    tags: _containers.RepeatedScalarFieldContainer[str]
    thumbnail: str
    sample: str
    labeling: str
    date: str
    author: str
    citation: str
    email: str
    map: _containers.ScalarMap[str, str]
    locdata_meta: _metadata_pb2.Metadata
    def __init__(
        self,
        format_version: _Optional[str] = ...,
        formats: _Optional[_Mapping[str, Format]] = ...,
        files: _Optional[_Iterable[_Union[FileInfo, _Mapping]]] = ...,
        name: _Optional[str] = ...,
        description: _Optional[str] = ...,
        tags: _Optional[_Iterable[str]] = ...,
        thumbnail: _Optional[str] = ...,
        sample: _Optional[str] = ...,
        labeling: _Optional[str] = ...,
        date: _Optional[str] = ...,
        author: _Optional[str] = ...,
        citation: _Optional[str] = ...,
        email: _Optional[str] = ...,
        map: _Optional[_Mapping[str, str]] = ...,
        locdata_meta: _Optional[_Union[_metadata_pb2.Metadata, _Mapping]] = ...,
    ) -> None: ...



from importlib.metadata import version

import numpy as np
import pytest

from locan.dependencies import HAS_DEPENDENCY
from locan.process.register import (
    _register_cc_picasso,
    _register_cc_skimage,
    register_cc,
    register_icp,
)
from locan.process.transform import transform_affine
from locan.tests.conftest import get_open3d_version


@pytest.mark.skipif(not HAS_DEPENDENCY["open3d"], reason="Test requires open3d.")
@pytest.mark.skipif(
    HAS_DEPENDENCY["open3d"]
    and get_open3d_version().startswith("0.18")
    and version("numpy").startswith("2"),
    reason="Test requires open3d>0.18 or numpy<2.",
)
def test_register_icp_2d(locdata_blobs_2d):
    locdata_2d_transformed = transform_affine(locdata_blobs_2d, offset=(100, 100))
    offset_target = np.array([100.0, 100.0])
    matrix_target = np.array([[1, 0], [0, 1]])

    matrix, offset = register_icp(
        locdata_blobs_2d, locdata_2d_transformed, verbose=False
    )
    assert np.allclose(np.array(offset), offset_target)
    assert np.allclose(matrix, matrix_target)

    matrix, offset = register_icp(
        locdata_blobs_2d,
        locdata_2d_transformed,
        pre_translation=(-90, -90),
        verbose=False,
    )
    offset_target = np.array([10.0, 10.0])
    assert np.allclose(np.array(offset), offset_target)
    assert np.allclose(matrix, matrix_target)


@pytest.mark.skipif(not HAS_DEPENDENCY["open3d"], reason="Test requires open3d.")
@pytest.mark.skipif(
    HAS_DEPENDENCY["open3d"]
    and get_open3d_version().startswith("0.18")
    and version("numpy").startswith("2"),
    reason="Test requires open3d>0.18 or numpy<2.",
)
def test_register_icp_3d(locdata_blobs_3d):
    locdata_3d_transformed = transform_affine(locdata_blobs_3d, offset=(100, 100, 100))
    offset_target = np.array([100.0, 100.0, 100.0])
    matrix_target = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])

    matrix, offset = register_icp(
        locdata_blobs_3d, locdata_3d_transformed, verbose=False
    )
    assert np.allclose(np.array(offset), offset_target)
    assert np.allclose(matrix, matrix_target)

    matrix, offset = register_icp(
        locdata_blobs_3d,
        locdata_3d_transformed,
        pre_translation=(-90, -90, -90),
        verbose=False,
    )
    offset_target = np.array([10.0, 10.0, 10.0])
    assert np.allclose(np.array(offset), offset_target)
    assert np.allclose(matrix, matrix_target)


def test__register_cc_picasso(locdata_blobs_2d):
    locdata_2d_transformed = transform_affine(locdata_blobs_2d, offset=(100, 50))

    offset_target = np.array([100.0, 50.0])
    matrix_target = np.array([[1, 0], [0, 1]])

    matrix, offset = _register_cc_picasso(
        locdata_blobs_2d, locdata_2d_transformed, bin_size=50, verbose=False
    )
    assert np.allclose(np.array(offset), offset_target, atol=5)
    assert np.allclose(matrix, matrix_target)

    matrix, offset = _register_cc_picasso(
        locdata_blobs_2d, locdata_2d_transformed, bin_size=(10, 50), verbose=False
    )
    assert np.allclose(np.array(offset), offset_target, atol=5)
    assert np.allclose(matrix, matrix_target)


def test__register_cc_skimage(locdata_blobs_2d):
    locdata_2d_transformed = transform_affine(locdata_blobs_2d, offset=(100, 50))

    offset_target = np.array([100.0, 50.0])
    matrix_target = np.array([[1, 0], [0, 1]])

    matrix, offset = _register_cc_skimage(
        locdata_blobs_2d, locdata_2d_transformed, bin_size=50
    )
    assert np.allclose(np.array(offset), offset_target, atol=5)
    assert np.allclose(matrix, matrix_target)

    matrix, offset = _register_cc_skimage(
        locdata_blobs_2d, locdata_2d_transformed, bin_size=(10, 50)
    )
    assert np.allclose(np.array(offset), offset_target, atol=5)
    assert np.allclose(matrix, matrix_target)


def test_register_cc(locdata_blobs_2d):
    locdata_2d_transformed = transform_affine(locdata_blobs_2d, offset=(100, 50))

    offset_target = np.array([100.0, 50.0])
    matrix_target = np.array([[1, 0], [0, 1]])

    matrix, offset = register_cc(locdata_blobs_2d, locdata_2d_transformed, bin_size=50)
    assert np.allclose(np.array(offset), offset_target, atol=5)
    assert np.allclose(matrix, matrix_target)

    matrix, offset = register_cc(
        locdata_blobs_2d, locdata_2d_transformed, bin_size=(10, 50)
    )
    assert np.allclose(np.array(offset), offset_target, atol=5)
    assert np.allclose(matrix, matrix_target)



"""

Transform localization intensities.

Localization intensities can be given in counts, electrons or photons.
This module provides often used transformation functions.

"""

from __future__ import annotations

import copy
import logging
import sys

import numpy as np
import numpy.typing as npt

from locan.data import metadata_pb2
from locan.data.locdata import LocData
from locan.data.metadata_utils import _modify_meta

__all__: list[str] = ["transform_counts_to_photons"]

logger = logging.getLogger(__name__)


def _transform_counts_to_photons(
    intensities: npt.ArrayLike,
    offset: int | float = 0,
    gain: int | float = 1,
    electrons_per_count: int | float = 1,
) -> npt.NDArray[np.float64]:
    """
    Convert camera analog-to-digital converter (ADC) counts into photo
    electrons.
    Quantum efficiency at the detected wavelength and collection efficiency
    of the optical system are not taken into account here.

    Parameters
    ----------
    intensities
        Intensity values in ADC counts.
    offset
        Camera offset
    gain
        Camera gain
    electrons_per_count
        Conversion factor for photo electrons per ADC count.

    Returns
    -------
    npt.NDArray[np.float64]
        Photo electrons
    """
    intensities = np.asarray(intensities, dtype=np.float64)
    return (intensities - offset) * electrons_per_count / gain  # type: ignore


def transform_counts_to_photons(
    locdata: LocData,
    loc_properties: str | list[str] | None = None,
    metadata: metadata_pb2.Camera | None = None,
) -> LocData:
    """
    Convert camera analog-to-digital converter (ADC) counts into photo
    electrons.
    Quantum efficiency at the detected wavelength and collection efficiency
    of the optical system are not taken into account here.

    Parameters
    ----------
    locdata
        Localization data on which to perform the manipulation.
    loc_properties
        Localization properties to be converted.
        If None the `intensity` values of locdata are used.
    metadata
        Camera metadata with attribute offset, gain, electrons_per_count
        If None, locdata.meta.experiment.setups[0].optical_units[0].detection.camera
        is used.

    Returns
    -------
    LocData
        Localization data with converted intensity values.
    """
    local_parameter = locals()

    if not len(locdata):
        logger.warning("Locdata is empty.")
        return locdata

    if loc_properties is None:
        loc_properties = ["intensity"]
    elif isinstance(loc_properties, str):
        loc_properties = [loc_properties]

    # instantiate
    # todo: possibly add inplace keyword
    new_locdata = copy.copy(locdata)

    loc_properties_converted = []
    for loc_property in loc_properties:
        if loc_property not in locdata.data.columns:
            logger.warning(f"Localization property {loc_property} is not available.")
            continue

        try:
            index = [prob.name for prob in locdata.meta.localization_properties].index(
                loc_property
            )
            unit = locdata.meta.localization_properties[index].unit
        except ValueError:
            index = None
            unit = None

        if index is not None:
            if locdata.meta.localization_properties[index].unit == "photons":
                logger.warning(
                    f"Localization property {loc_property} is already provided with "
                    f"unit photons"
                )
                continue

        intensities = getattr(locdata.data, loc_property)

        if metadata is None:
            metadata = (
                locdata.meta.experiment.setups[0].optical_units[0].detection.camera
            )

        offset = metadata.offset
        gain = metadata.gain
        electrons_per_count = metadata.electrons_per_count

        new_intensities = _transform_counts_to_photons(
            intensities,
            offset=offset,
            gain=gain,
            electrons_per_count=electrons_per_count,
        )

        if index is None:
            new_locdata.meta.localization_properties.add()
            new_locdata.meta.localization_properties[-1].name = loc_property
            new_locdata.meta.localization_properties[-1].unit = "photons"
            new_locdata.meta.localization_properties[-1].type = "float"
        else:
            new_locdata.meta.localization_properties[index].unit = "photons"
            new_locdata.meta.localization_properties[index].type = "float"

        assert new_locdata.dataframe is not None  # type narrowing # noqa: S101
        dataframe = new_locdata.dataframe.assign(**{loc_property: new_intensities})
        new_locdata = new_locdata.update(dataframe=dataframe)
        loc_properties_converted.append(loc_property)

    # update metadata
    meta_ = _modify_meta(
        locdata,
        new_locdata,
        function_name=sys._getframe().f_code.co_name,
        parameter=local_parameter,
        meta=new_locdata.meta,
    )
    new_locdata.meta = meta_

    logger.info(f"Successfully converted: {loc_properties_converted}.")
    return new_locdata



import numpy as np
import pandas as pd
import pytest

from locan.data.locdata_utils import (
    _bump_property_key,
    _check_loc_properties,
    _get_linked_coordinates,
    _get_loc_property_key_per_dimension,
)


@pytest.fixture()
def df_only_coordinates():
    return pd.DataFrame.from_dict(
        {"position_y": np.arange(3), "position_z": np.arange(3)}
    )


@pytest.fixture()
def df_with_zero_uncertainty():
    return pd.DataFrame.from_dict(
        {
            "position_y": np.arange(3),
            "uncertainty": np.arange(3),
            "position_z": np.arange(3),
            "uncertainty_z": np.arange(3),
        }
    )


@pytest.fixture()
def df_with_uncertainty():
    return pd.DataFrame.from_dict(
        {
            "position_y": np.arange(1, 3),
            "uncertainty": np.arange(1, 3),
            "position_z": np.arange(1, 3),
            "uncertainty_z": np.arange(1, 3),
        }
    )


@pytest.fixture()
def df_empty():
    return pd.DataFrame()


@pytest.fixture()
def df_single():
    return pd.DataFrame.from_dict(
        {
            "position_y": [1],
            "position_z": [2],
            "uncertainty_z": [5],
        }
    )


def test__get_loc_property_key_per_dimension(df_with_uncertainty):
    results = _get_loc_property_key_per_dimension(
        locdata=df_with_uncertainty, property_key="position"
    )
    assert results == [None, "position_y", "position_z"]

    results = _get_loc_property_key_per_dimension(
        locdata=df_with_uncertainty, property_key="uncertainty"
    )
    assert results == ["uncertainty", "uncertainty", "uncertainty_z"]

    results = _get_loc_property_key_per_dimension(
        locdata=df_with_uncertainty, property_key="not_existing"
    )
    assert results == [None, None, None]


def test__get_linked_coordinates(
    df_only_coordinates,
    df_with_zero_uncertainty,
    df_with_uncertainty,
    df_empty,
    df_single,
    caplog,
):
    results = _get_linked_coordinates(locdata=df_only_coordinates)
    assert results == {
        "position_y": 1.0,
        "uncertainty_y": 0.5773502691896257,
        "position_z": 1.0,
        "uncertainty_z": 0.5773502691896257,
    }

    results = _get_linked_coordinates(locdata=df_with_zero_uncertainty)
    assert all(np.isnan(results[key]) for key in results.keys())
    assert caplog.record_tuples[-1] == (
        "locan.data.locdata_utils",
        30,
        "Zero uncertainties occurred resulting in nan for weighted_mean and weighted_variance.",
    )

    results = _get_linked_coordinates(locdata=df_with_uncertainty)
    assert results == {
        "position_y": 1.2,
        "uncertainty_y": 0.32,
        "position_z": 1.2,
        "uncertainty_z": 0.32,
    }

    results = _get_linked_coordinates(
        locdata=df_only_coordinates, coordinate_keys="position_y"
    )
    assert results == {"position_y": 1.0, "uncertainty_y": 0.5773502691896257}

    results = _get_linked_coordinates(
        locdata=df_with_uncertainty, coordinate_keys="position_y"
    )
    assert results == {
        "position_y": 1.2,
        "uncertainty_y": 0.32,
    }

    results = _get_linked_coordinates(
        locdata=df_only_coordinates, coordinate_keys="position_z"
    )
    assert results == {"position_z": 1.0, "uncertainty_z": 0.5773502691896257}

    results = _get_linked_coordinates(
        locdata=df_with_uncertainty, coordinate_keys="position_z"
    )
    assert results == {
        "position_z": 1.2,
        "uncertainty_z": 0.32,
    }

    results = _get_linked_coordinates(locdata=df_empty)
    assert results == {}

    results = _get_linked_coordinates(locdata=df_single)
    assert results == {
        "position_y": 1,
        "uncertainty_y": 0,
        "position_z": 2,
        "uncertainty_z": 5,
    }


def test__bump_property_label():
    new_property = _bump_property_key(
        loc_properties=["test", "other_test"],
        loc_property="test",
        extension="_extended",
    )
    assert new_property == "test_extended"

    new_property = _bump_property_key(
        loc_properties=["test", "test_0"],
        loc_property="test",
    )
    assert new_property == "test_0_0"


def test__check_loc_properties(locdata_2d):
    with pytest.raises(ValueError):
        _check_loc_properties(locdata=locdata_2d, loc_properties=["test"])

    result = _check_loc_properties(locdata=locdata_2d, loc_properties="position_x")
    assert result == ["position_x"]

    result = _check_loc_properties(locdata=locdata_2d, loc_properties=None)
    assert result == ["position_x", "position_y"]

    result = _check_loc_properties(
        locdata=locdata_2d, loc_properties=locdata_2d.data.columns
    )
    assert result == locdata_2d.data.columns.tolist()



import numpy as np
import pandas as pd
import pytest

from locan import LocData
from locan.process.properties import range_from_collection, ranges, statistics


@pytest.fixture()
def locdata_simple():
    dict_ = {"position_x": [0, 0, 1, 4, 5], "position_y": [0, 1, 3, 4, 1]}
    return LocData(dataframe=pd.DataFrame.from_dict(dict_))


def test_statistics(locdata_simple):
    stat = statistics(locdata=locdata_simple)
    assert stat == {
        "position_x_count": 5.0,
        "position_x_min": 0.0,
        "position_x_max": 5.0,
        "position_x_mean": 2.0,
        "position_x_median": 1.0,
        "position_x_std": 2.3452078799117149,
        "position_x_sem": 1.0488088481701516,
        "position_y_count": 5.0,
        "position_y_min": 0.0,
        "position_y_max": 4.0,
        "position_y_mean": 1.8,
        "position_y_median": 1.0,
        "position_y_std": 1.6431676725154984,
        "position_y_sem": 0.73484692283495345,
    }


def test_statistics_from_dataframe(locdata_simple):
    stat = statistics(locdata=locdata_simple.data)
    assert stat == {
        "position_x_count": 5.0,
        "position_x_min": 0.0,
        "position_x_max": 5.0,
        "position_x_mean": 2.0,
        "position_x_median": 1.0,
        "position_x_std": 2.3452078799117149,
        "position_x_sem": 1.0488088481701516,
        "position_y_count": 5.0,
        "position_y_min": 0.0,
        "position_y_max": 4.0,
        "position_y_mean": 1.8,
        "position_y_median": 1.0,
        "position_y_std": 1.6431676725154984,
        "position_y_sem": 0.73484692283495345,
    }


def test_statistics_with_one_statfunction(locdata_simple):
    stat = statistics(locdata=locdata_simple, statistic_keys="min")
    assert stat == {"position_x_min": 0, "position_y_min": 0}


def test_statistics_from_Series(locdata_simple):
    stat = statistics(locdata=locdata_simple.data["position_x"], statistic_keys="mean")
    assert stat["position_x_mean"] == 2
    stat = statistics(
        locdata=locdata_simple.data["position_x"], statistic_keys=("min", "max")
    )
    assert stat == {"position_x_min": 0, "position_x_max": 5}


def test_statistics_with_new_column(locdata_simple):
    locdata_simple.dataframe = locdata_simple.dataframe.assign(new=np.arange(5))
    stat = statistics(locdata=locdata_simple)
    # print(stat)
    assert stat == {
        "position_x_count": 5.0,
        "position_x_min": 0.0,
        "position_x_max": 5.0,
        "position_x_mean": 2.0,
        "position_x_median": 1.0,
        "position_x_std": 2.3452078799117149,
        "position_x_sem": 1.0488088481701516,
        "position_y_count": 5.0,
        "position_y_min": 0.0,
        "position_y_max": 4.0,
        "position_y_mean": 1.8,
        "position_y_median": 1.0,
        "position_y_std": 1.6431676725154984,
        "position_y_sem": 0.73484692283495345,
        "new_count": 5.0,
        "new_min": 0.0,
        "new_max": 4.0,
        "new_mean": 2.0,
        "new_median": 2.0,
        "new_std": 1.5811388300841898,
        "new_sem": 0.70710678118654757,
    }


def test_range_from_collection(locdata_3d):
    collection = LocData.from_chunks(locdata_3d, n_chunks=2)
    result = range_from_collection(collection.references)
    assert result.position_x.min == 1
    assert result.position_x.max == 5
    assert result.position_z.min == 1
    assert result.position_z.max == 5

    result = range_from_collection(collection.references, loc_properties=["position_x"])
    assert result.position_x.min == 1
    assert result.position_x.max == 5

    result = range_from_collection(collection.references, loc_properties="position_x")
    assert result.position_x.min == 1
    assert result.position_x.max == 5

    result = range_from_collection(collection.references, loc_properties=True)
    assert result.position_x.min == 1
    assert result.position_x.max == 5
    assert result.frame.min == 1


def test__ranges(locdata_2d):
    ranges_ = ranges(locdata_2d)
    assert np.array_equal(ranges_, [[1, 5], [1, 6]])
    ranges_ = ranges(locdata_2d, loc_properties=True)
    assert np.array_equal(ranges_, [[1, 5], [1, 6], [1, 6], [80, 150]])
    ranges_ = ranges(locdata_2d, loc_properties="intensity")
    assert np.array_equal(ranges_, [[80, 150]])
    ranges_ = ranges(locdata_2d, loc_properties=["intensity"])
    assert np.array_equal(ranges_, [[80, 150]])
    ranges_ = ranges(locdata_2d, loc_properties=["frame", "intensity"])
    assert np.array_equal(ranges_, [[1, 6], [80, 150]])
    ranges_ = ranges(locdata_2d, loc_properties=("frame", "intensity"))
    assert np.array_equal(ranges_, [[1, 6], [80, 150]])
    ranges_ = ranges(locdata_2d, loc_properties=None, special="zero")
    assert np.array_equal(ranges_, [[0, 5], [0, 6]])
    ranges_ = ranges(locdata_2d, loc_properties=True, special="link")
    assert np.array_equal(ranges_, [[1, 150], [1, 150], [1, 150], [1, 150]])


def test__ranges_empty(locdata_empty, locdata_single_localization):
    assert ranges(locdata_empty) is None
    ranges_ = ranges(locdata_single_localization, loc_properties=None)
    assert np.array_equal(ranges_, [[1, 2], [1, 2]])
    ranges_ = ranges(locdata_single_localization, loc_properties=True)
    assert np.array_equal(ranges_, [[1, 2], [1, 2], [1, 2], [1, 2]])



"""
Base class that serves as template for a specialized analysis class.

It also provides helper functions to be used in specialized analysis classes.
"""

from __future__ import annotations

import sys
from copy import copy
from typing import Any

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

from google.protobuf import json_format
from scipy import stats

from locan.analysis import metadata_analysis_pb2


class _Analysis:
    """
    Base class for standard analysis procedures.

    Parameters
    ----------
    meta : metadata_analysis_pb2.AMetadata | dict[str, Any] | None
        Metadata about the current analysis routine.
    kwargs
        Parameter that are passed to the algorithm.

    Attributes
    ----------
    locdata : LocData
        Localization data.
    parameter : dict[str, Any]
        A dictionary with all settings (i.e. the kwargs) for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata | None
        Metadata about the current analysis routine.
    results : Any | None
        Computed results.
    """

    count: int = 0
    """A counter for counting Analysis class instantiations (class attribute)."""

    def __init__(
        self,
        meta: metadata_analysis_pb2.AMetadata | dict[str, Any] | None,
        **kwargs: Any,
    ) -> None:
        self.__class__.count += 1

        self.parameter: dict[str, Any] = kwargs
        self.meta: metadata_analysis_pb2.AMetadata = self._init_meta()
        self.meta = self._update_meta(meta)
        self.results: Any | None = None

    @staticmethod
    def _get_parameters(locals: dict[str, Any]) -> dict[str, Any]:
        """Get parameters from curated locals."""
        parameters = copy(locals)
        for key in ["self", "__class__"]:
            parameters.pop(key)
        if "kwargs" in parameters:
            parameters.update(parameters["kwargs"])
            parameters.pop("kwargs")
        return parameters

    def __del__(self) -> None:
        """Update the counter upon deletion of class instance."""
        self.__class__.count -= 1

    def __repr__(self) -> str:
        """Return representation of the Analysis class."""
        parameter_string = ", ".join(
            (f"{key}={value}" for key, value in self.parameter.items())
        )
        return f"{self.__class__.__name__}({parameter_string})"

    def __getstate__(self) -> dict[str, Any]:
        """Modify pickling behavior."""
        # Copy the object's state from self.__dict__ to avoid modifying the original state.
        state = self.__dict__.copy()
        # Serialize the unpicklable protobuf entries.
        json_string = json_format.MessageToJson(
            self.meta, always_print_fields_with_no_presence=False
        )
        state["meta"] = json_string
        return state

    def __setstate__(self, state: dict[str, Any]) -> None:
        """Modify pickling behavior."""
        # Restore instance attributes.
        self.__dict__.update(state)
        # Restore protobuf class for meta attribute
        self.meta = metadata_analysis_pb2.AMetadata()
        self.meta = json_format.Parse(state["meta"], self.meta)

    def __bool__(self) -> bool:
        if self.results is not None:
            return True
        else:
            return False

    def compute(self, *args: Any, **kwargs: Any) -> Self:
        """
        Apply analysis routine with the specified parameters on locdata and
        return results.
        """
        raise NotImplementedError

    def report(self, *args: Any, **kwargs: Any) -> Any:
        """Show a report about analysis results."""
        raise NotImplementedError

    def _init_meta(self) -> metadata_analysis_pb2.AMetadata:
        meta_ = metadata_analysis_pb2.AMetadata()
        meta_.identifier = str(self.__class__.count)
        meta_.creation_time.GetCurrentTime()
        meta_.method.name = str(self.__class__.__name__)
        meta_.method.parameter = str(self.parameter)
        return meta_

    def _update_meta(
        self, meta: metadata_analysis_pb2.AMetadata | dict[str, Any] | None = None
    ) -> metadata_analysis_pb2.AMetadata:
        meta_ = self.meta
        if meta is None:
            pass
        else:
            try:
                meta_.MergeFrom(meta)  # type: ignore[arg-type]
            except TypeError:
                for key, value in meta.items():  # type: ignore
                    setattr(meta_, key, value)

        return meta_


# Dealing with scipy.stats


def _list_parameters(distribution: str | stats.rv_continuous) -> list[str]:
    """
    List parameters for scipy.stats.rv_continuous.

    Parameters
    ----------
    distribution
        Distribution of choice.

    Returns
    -------
    list[str]
        A list of distribution parameter strings.
    """
    if isinstance(distribution, str):
        distribution = getattr(stats, distribution)
    if distribution.shapes:
        parameters = [name.strip() for name in distribution.shapes.split(",")]
    else:
        parameters = []
    if distribution.name in stats._discrete_distns._distn_names:
        parameters += ["loc"]
    elif distribution.name in stats._continuous_distns._distn_names:
        parameters += ["loc", "scale"]
    else:
        raise TypeError("Distribution name not found in discrete or continuous lists.")

    return parameters



"""
Localization data.

This module contains data models to deal with single-molecule localization data.

Submodules:
-----------

.. autosummary::
   :toctree: ./

   hulls
   locdata
   metadata_utils
   regions
   validation

"""

from __future__ import annotations

from importlib import import_module

import locan.data.metadata_pb2

# explicit re-exports for nested submodules are required by myp
from locan.data.hulls import (
    AlphaComplex as AlphaComplex,
    AlphaShape as AlphaShape,
    BoundingBox as BoundingBox,
    ConvexHull as ConvexHull,
    OrientedBoundingBox as OrientedBoundingBox,
)
from locan.data.locdata import *
from locan.data.metadata_utils import *
from locan.data.regions import (
    AxisOrientedCuboid as AxisOrientedCuboid,
    AxisOrientedHypercuboid as AxisOrientedHypercuboid,
    Cuboid as Cuboid,
    Ellipse as Ellipse,
    EmptyRegion as EmptyRegion,
    Interval as Interval,
    MultiPolygon as MultiPolygon,
    Polygon as Polygon,
    Rectangle as Rectangle,
    Region as Region,
    Region1D as Region1D,
    Region2D as Region2D,
    Region3D as Region3D,
    RegionND as RegionND,
    RoiRegion as RoiRegion,
    expand_region as expand_region,
    regions_union as regions_union,
    surrounding_region as surrounding_region,
)
from locan.data.validation import *

submodules: list[str] = [
    "hulls",
    "locdata",
    "metadata_utils",
    "regions",
    "validation",
]

__all__: list[str] = []

for submodule in submodules:
    module_ = import_module(name=f".{submodule}", package="locan.data")
    __all__.extend(module_.__all__)



import pickle
import tempfile
from copy import deepcopy
from pathlib import Path

import pandas as pd
import pytest

from locan import Files


@pytest.fixture(scope="session")
def test_files(tmp_path_factory):
    directory = tmp_path_factory.mktemp("test_directory")
    (directory / "sub_directory").mkdir()
    files = [
        directory / "sub_directory" / "file_group_a_0.data",
        directory / "sub_directory" / "file_group_a_1.data",
        directory / "sub_directory" / "file_group_b_2.data",
        directory / "sub_directory" / "corresponding_file_0.data",
        directory / "metadata.meta",
    ]
    for file_ in files:
        file_.touch()
    return directory


def test_fixture_test_files(test_files):
    files = list(test_files.glob("**/*.*"))
    assert len(files) == 5


def test_Files_init(test_files, capfd):
    files = Files()
    assert files.directory is None
    assert "file_path" in files.df.columns
    assert isinstance(files.df, pd.DataFrame)
    assert len(files.df) == 0
    assert files.print_summary() is None

    with pytest.raises(ValueError):
        Files(directory=test_files / "does_not_exist")

    files = Files(
        directory=test_files,
        df={"my_file": [test_files / "sub_directory" / "file_group_a_0.data"]},
        column="my_file",
    )
    assert files.directory == test_files
    assert isinstance(files.df, pd.DataFrame)
    assert "my_file" in files.df.columns
    assert isinstance(files.df.my_file[0], Path)
    assert len(files.df) == 1
    assert files.print_summary() is None

    with pytest.raises(FileExistsError):
        Files(df={"file_path": [test_files / "does_not_exist.data"]})

    files = Files(df={"file_path": [test_files / "does_not_exist.data"]}, exists=False)
    assert files.directory is None
    assert "file_path" in files.df.columns
    assert isinstance(files.df, pd.DataFrame)
    assert isinstance(files.df.file_path[0], Path)
    assert len(files.df) == 1
    assert files.print_summary() is None

    capfd.readouterr()


def test_Files_from_path(test_files, capfd):
    file_path = test_files / "sub_directory" / "file_group_a_0.data"
    files = Files.from_path(file_path)
    assert files.directory is None
    assert isinstance(files.df, pd.DataFrame)
    assert isinstance(files.df.file_path[0], Path)
    assert len(files.df) == 1
    assert files.print_summary() is None

    files = Files.from_path(str(file_path))
    assert files.directory is None
    assert isinstance(files.df, pd.DataFrame)
    assert isinstance(files.df.file_path[0], Path)
    assert len(files.df) == 1
    assert files.print_summary() is None

    file_path = [
        test_files / "sub_directory" / "file_group_a_0.data",
        test_files / "sub_directory" / "file_group_a_1.data",
    ]

    files = Files.from_path(file_path)
    assert files.directory is None
    assert isinstance(files.df, pd.DataFrame)
    assert isinstance(files.df.file_path[0], Path)
    assert len(files.df) == 2
    assert files.print_summary() is None

    files = Files.from_path([str(f_) for f_ in file_path])
    assert files.directory is None
    assert isinstance(files.df, pd.DataFrame)
    assert isinstance(files.df.file_path[0], Path)
    assert len(files.df) == 2
    assert files.print_summary() is None

    capfd.readouterr()


def test_Files_indexing(test_files, capfd):
    file_path = [
        test_files / "sub_directory" / "file_group_a_0.data",
        test_files / "sub_directory" / "file_group_a_1.data",
    ]

    files = Files.from_path(file_path)
    # test __iter__:
    for file_ in files:
        assert isinstance(file_, tuple)
        assert file_.file_path

    # test __getitem__:
    file_ = files[0]
    assert isinstance(file_, pd.Series)
    assert file_.file_path

    file_ = files[0:2]
    assert isinstance(file_, Files)
    assert len(file_.df.file_path) == 2

    file_ = files[0:1]
    assert isinstance(file_, Files)
    assert len(file_.df.file_path) == 1

    file_ = files[[0]]
    assert isinstance(file_, Files)
    assert len(file_.df.file_path) == 1


def test_Files_from_glob(test_files, capfd):
    files = Files().from_glob(pattern="this_should_not_ever_exist_#")
    assert files.directory.resolve() == Path().cwd().resolve()
    assert files.df.empty

    files = Files().from_glob(directory=test_files)
    assert files.directory.resolve() == test_files.resolve()
    assert files.df.empty

    files = Files().from_glob(directory=test_files, pattern="**/*.data")
    assert files.directory.resolve() == test_files.resolve()
    assert len(files.df) == 4

    files = Files().from_glob(
        directory=test_files, pattern="**/*.data", regex="corresponding"
    )
    assert files.directory.resolve() == test_files.resolve()
    assert len(files.df) == 1


def test_Files_add_glob(test_files, capfd, caplog):
    files = Files().from_glob(directory=test_files, pattern="**/*file_group_a_0.*")
    files.add_glob(pattern="**/*_0.*", regex="corresponding")
    assert len(files.df) == 1
    assert "file_path" in files.df.columns
    assert "other_file_path" in files.df.columns

    files = Files().from_glob(directory=test_files, pattern="**/*file_group_a*.*")
    files.add_glob(pattern="**/*_0.*", regex="corresponding")
    assert len(files.df) == 2
    assert "file_path" in files.df.columns
    assert "other_file_path" in files.df.columns
    assert caplog.record_tuples == [
        ("locan.locan_io.files", 30, "Not all files are matched.")
    ]


def test_Files_match_files(test_files, capfd, caplog):
    files = Files().from_glob(directory=test_files, pattern="**/*file_group_a_0.*")
    files_match = Files().from_glob(
        directory=test_files, pattern="**/*_0.*", regex="corresponding"
    )
    files.match_files(files=files_match.df["file_path"])
    assert len(files.df) == 1
    assert "file_path" in files.df.columns
    assert "other_file_path" in files.df.columns

    files = Files().from_glob(directory=test_files, pattern="**/*file_group_a*.*")
    files_match = Files().from_glob(
        directory=test_files, pattern="**/*_0.*", regex="corresponding"
    )
    files.match_files(files=files_match.df["file_path"])
    assert len(files.df) == 2
    assert "file_path" in files.df.columns
    assert "other_file_path" in files.df.columns
    assert caplog.record_tuples == [
        ("locan.locan_io.files", 30, "Not all files are matched.")
    ]


def test_Files_exclude(test_files, capfd):
    files = Files().from_glob(directory=test_files, pattern="**/*.*")
    assert len(files.df) == 5

    files_new = deepcopy(files)
    stoplist = None
    files_new = files_new.exclude(stoplist=stoplist)
    assert len(files_new.df) == len(files.df)
    assert "file_path" in files_new.df.columns

    files_new = deepcopy(files)
    stoplist = []
    files_new = files_new.exclude(stoplist=stoplist)
    assert len(files_new.df) == len(files.df)
    assert "file_path" in files_new.df.columns

    files_new = deepcopy(files)
    stoplist = Files()
    files_new = files_new.exclude(stoplist=stoplist)
    assert len(files_new.df) == len(files.df)
    assert "file_path" in files_new.df.columns

    files_new = deepcopy(files)
    stoplist = [False, True, True, True, True]
    files_new = files_new.exclude(stoplist=stoplist)
    assert len(files_new.df) == 1
    assert "file_path" in files_new.df.columns

    with pytest.raises(ValueError):
        stoplist = [False, True]
        files_new.exclude(stoplist=stoplist)

    files_new = deepcopy(files)
    stoplist = Files().from_glob(directory=test_files, pattern="**/*.data")
    stoplist = list(stoplist.df.file_path)
    files_new = files_new.exclude(stoplist=stoplist)
    assert len(files_new.df) == 1

    files_new = deepcopy(files)
    stoplist = Files().from_glob(directory=test_files, pattern="**/*.data")
    stoplist = list(stoplist.df.file_path.astype("string"))
    files_new = files_new.exclude(stoplist=stoplist)
    assert len(files_new.df) == 1

    files_new = deepcopy(files)
    stoplist = Files().from_glob(directory=test_files, pattern="**/*.data")
    files_new = files_new.exclude(stoplist=stoplist)
    assert len(files_new.df) == 1

    files_new = deepcopy(files)
    stoplist = Files().from_glob(
        directory=test_files, pattern="**/*.data", column="stoplist"
    )
    files_new = files_new.exclude(stoplist=stoplist, column_stoplist="stoplist")
    assert len(files_new.df) == 1

    files_new = deepcopy(files)
    files_new = files_new.exclude(stoplist=files)
    assert len(files_new.df) == 0


def test_Files_concatenate(test_files, capfd):
    files_1 = Files().from_glob(directory=test_files, pattern="**/*.data", column="one")
    files_2 = Files().from_glob(directory=test_files, pattern="**/*.meta", column="one")
    files = Files().concatenate(files=[files_1, files_2], exists=False)
    assert len(files.df) == 5
    assert "one" in files.df.columns


def test_Files_match_file_upstream(test_files, capfd):
    files = Files().from_glob(directory=test_files, pattern="**/*.data")
    files_new = files.match_file_upstream(
        pattern="*.meta", other_column="matching_file"
    )
    assert len(files_new.df) == 4
    assert len(files_new.df.columns) == 2

    files = Files().from_glob(directory=test_files, pattern="**/*.data")
    files_new = files.match_file_upstream(
        pattern="*.data", other_column="matching_file", regex="corresponding"
    )
    assert len(files_new.df) == 4
    assert len(files_new.df.columns) == 2
    assert files_new.df.matching_file.notna().any()


def test_set_group_identifier(test_files, capfd):
    files = Files().from_glob(directory=test_files, pattern="**/*.*")
    # print(files.files.applymap(lambda x: x.name))
    files.set_group_identifier(name="group_subdir", pattern="sub_directory")
    assert files.df.group[files.df.group == "group_subdir"].count() == 4
    assert files.df.group.dtype == "category"
    files.set_group_identifier(name="group_data", glob="*.data")
    assert files.df.group[files.df.group == "group_data"].count() == 4
    assert files.df.group.dtype == "category"
    files.set_group_identifier(name="group_b", regex="group_b")
    assert files.df.group[files.df.group == "group_b"].count() == 1
    assert files.df.group.dtype == "category"
    files = Files().from_glob(directory=test_files, pattern="**/*.*")
    files.set_group_identifier(
        name="group_b", pattern="sub_directory", glob="*.txt", regex="group_b"
    )
    assert files.df.group[files.df.group == "group_b"].count() == 0
    assert files.df.group.dtype == "category"
    files.set_group_identifier(
        name="group_b", pattern="sub_directory", glob="*.data", regex="group_b"
    )
    assert files.df.group[files.df.group == "group_b"].count() == 1
    assert files.df.group.dtype == "category"

    assert "group_b" in files.group_identifiers()
    assert isinstance(files.grouped(), pd.core.groupby.DataFrameGroupBy)


def test_pickling_Files(test_files):
    files = Files().from_glob(directory=test_files, pattern="**/*.*")
    assert len(files.df) == 5

    with tempfile.TemporaryDirectory() as tmp_directory:
        file_path = Path(tmp_directory) / "pickled_files.pickle"
        with open(file_path, "wb") as file:
            pickle.dump(files, file, pickle.HIGHEST_PROTOCOL)
        with open(file_path, "rb") as file:
            pickled_files = pickle.load(file)  # noqa S301
        assert len(pickled_files.df) == len(files.df)



"""

Track localizations.

This module provides functions for tracking localizations
(i.e. clustering localization data in time).
The functions take LocData as input and compute new LocData objects.
It makes use of the trackpy package.

"""

from __future__ import annotations

import sys
from typing import Any

import pandas as pd

from locan.data.locdata import LocData
from locan.dependencies import HAS_DEPENDENCY, needs_package

if HAS_DEPENDENCY["trackpy"]:
    from trackpy import link_df


__all__: list[str] = ["link_locdata", "track"]


@needs_package("trackpy")
def link_locdata(
    locdata: LocData,
    search_range: float | tuple[float, ...] = 40,
    memory: int = 0,
    **kwargs: Any,
) -> pd.Series[Any]:
    """
    Track localizations, i.e. cluster localizations in time when nearby in
    successive frames.
    This function applies the trackpy linking method to LocData objects.

    Parameters
    ----------
    locdata
        Localization data on which to perform the manipulation.
    search_range
        The maximum distance features can move between frames,
        optionally per dimension
    memory
        The maximum number of frames during which a feature can vanish,
        then reappear nearby, and be considered the same particle.
    kwargs
        Other parameters passed to trackpy.link_df().

    Returns
    -------
    pandas.Series[Any]
        A series named 'Track' referring to the track number.

    Note
    ----
    In order to switch off the printout from :func:`trackpy.link` and increase
    performance use :func:`trackpy.quiet()` to silence the logging outputs.
    """
    df = link_df(
        locdata.data,
        search_range=search_range,
        memory=memory,
        pos_columns=locdata.coordinate_keys,
        t_column="frame",
        **kwargs,
    )
    return_series: pd.Series[Any] = df["particle"]
    return_series.name = "track"
    return return_series


def track(
    locdata: LocData,
    search_range: float | tuple[float, ...] = 40,
    memory: int = 0,
    **kwargs: Any,
) -> tuple[LocData, pd.Series[Any]]:
    """
    Cluster (in time) localizations in LocData that are nearby in successive
    frames. Clustered localizations are identified by the trackpy linking
    method.

    The new locdata object carries properties with the same name as the
    original `locata`.
    They are computed as sum for `intensity`, as the first value for `frame`,
    and as mean for all other properties.

    Parameters
    ----------
    locdata
        Localization data on which to perform the manipulation.
    search_range
        The maximum distance features can move between frames, optionally per
        dimension
    memory
        The maximum number of frames during which a feature can vanish, then
        reappear nearby, and be considered the same particle.
    kwargs
        Other parameters passed to trackpy.link_df.

    Returns
    -------
    tuple[Locdata, pandas.Series[Any]]
        A new LocData instance assembling all generated selections
        (i.e. localization cluster).
        A series named 'Track' referring to the track number.

    Note
    ----
    In order to switch off the printout from :func:`trackpy.link` and increase
    performance use :func:`trackpy.quiet()` to silence the logging outputs.
    """
    parameter = locals()

    track_series = link_locdata(locdata, search_range, memory, **kwargs)
    grouped = track_series.groupby(track_series)
    selections = [
        LocData.from_selection(locdata=locdata, indices=group.index.values)
        for _, group in grouped
    ]
    collection = LocData.from_collection(selections)

    # metadata
    del locdata.meta.history[:]
    locdata.meta.history.add(
        name=sys._getframe().f_code.co_name, parameter=str(parameter)
    )

    return collection, track_series



from copy import deepcopy

import boost_histogram as bh
import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pytest

from locan import LocData
from locan.analysis import ConvexHullExpectation, ConvexHullExpectationBatch
from locan.analysis.convex_hull_expectation import (
    ConvexHullExpectationResource,
    ConvexHullExpectationResults,
    ConvexHullExpectationValues,
    ConvexHullProperty,
    _get_convex_hull_property_expectation,
    _get_resource,
    compute_convex_hull_region_measure_2d,
)


def test_get_resource():
    for convex_hull_property_item in ConvexHullProperty:
        resource_ = ConvexHullExpectationResource[convex_hull_property_item.name]
        result = _get_resource(
            resource_directory="locan.analysis.resources.convex_hull_expectation",
            resource=resource_,
        )
        assert isinstance(result, ConvexHullExpectationValues)


def test_compute_convex_hull_region_measure_2d():
    computed = compute_convex_hull_region_measure_2d(n_points=3)
    from_resource = _get_resource(
        resource_directory="locan.analysis.resources.convex_hull_expectation",
        resource=ConvexHullExpectationResource["REGION_MEASURE_2D"],
    ).expectation[0]
    assert pytest.approx(computed, rel=0.01) == from_resource == 0.87

    computed = compute_convex_hull_region_measure_2d(n_points=100)
    from_resource = _get_resource(
        resource_directory="locan.analysis.resources.convex_hull_expectation",
        resource=ConvexHullExpectationResource["REGION_MEASURE_2D"],
    ).expectation[97]
    assert pytest.approx(computed, rel=0.001) == from_resource

    computed = compute_convex_hull_region_measure_2d(n_points=2, sigma=2)
    assert np.isnan(computed)


def test__get_convex_hull_property_expectation_region_measure_2d():
    result = _get_convex_hull_property_expectation(
        convex_hull_property="region_measure_2d", n_points=10, sigma=2
    )
    assert result == ConvexHullExpectationValues(
        n_points=np.array([10]), expectation=21.52, std_pos=9.16, std_neg=6.72
    )

    result = _get_convex_hull_property_expectation(
        convex_hull_property="region_measure_2d", n_points=300, sigma=2
    )
    assert result.n_points.size == 0

    result = _get_convex_hull_property_expectation(
        convex_hull_property="region_measure_2d", n_points=[3, 10], sigma=2
    )
    assert np.array_equal(result.expectation, np.array([3.48, 21.52]))

    result = _get_convex_hull_property_expectation(
        convex_hull_property="region_measure_2d", n_points=[1, 3, 10], sigma=2
    )
    assert np.array_equal(result.n_points, np.array([3, 10]))
    assert np.array_equal(result.expectation, np.array([3.48, 21.52]))

    result = _get_convex_hull_property_expectation(
        convex_hull_property="region_measure_2d", n_points=np.array([1, 3, 10]), sigma=2
    )
    assert np.array_equal(result.n_points, np.array([3, 10]))
    assert np.array_equal(result.expectation, np.array([3.48, 21.52]))


def test__get_convex_hull_property_expectation_subregion_measure_2d():
    result = _get_convex_hull_property_expectation(
        convex_hull_property="subregion_measure_2d", n_points=10, sigma=2
    )
    assert result == ConvexHullExpectationValues(
        n_points=np.array([10]),
        expectation=np.array([19.34]),
        std_pos=np.array([3.62]),
        std_neg=np.array([3.26]),
    )

    result = _get_convex_hull_property_expectation(
        convex_hull_property="subregion_measure_2d", n_points=[3, 10], sigma=2
    )
    assert np.array_equal(result.expectation, np.array([10.64, 19.34]))


def test__get_convex_hull_property_expectation_region_measure_3d():
    result = _get_convex_hull_property_expectation(
        convex_hull_property="region_measure_3d", n_points=10, sigma=2
    )
    assert result == ConvexHullExpectationValues(
        n_points=np.array([10]),
        expectation=np.array([49.2]),
        std_pos=np.array([28.08]),
        std_neg=np.array([18.24]),
    )

    result = _get_convex_hull_property_expectation(
        convex_hull_property="region_measure_3d", n_points=[4, 10], sigma=2
    )
    assert np.array_equal(result.expectation, np.array([4.24, 49.2]))


def test__get_convex_hull_property_expectation_subregion_measure_3d():
    result = _get_convex_hull_property_expectation(
        convex_hull_property="subregion_measure_3d", n_points=10, sigma=2
    )
    assert result == ConvexHullExpectationValues(
        n_points=np.array([10]), expectation=86.04, std_pos=27.72, std_neg=21.8
    )

    result = _get_convex_hull_property_expectation(
        convex_hull_property="subregion_measure_3d", n_points=[4, 10], sigma=2
    )
    assert np.array_equal(result.expectation, np.array([27.72, 86.04]))


def test_enums():
    for member in ConvexHullProperty:
        assert ConvexHullExpectationResource[member.name]


class TestConvexHullPropertyExpectation:
    def test_init(self, locdata_2d):
        cpe = ConvexHullExpectation(meta={"comment": "this is an example"})
        assert (
            str(cpe) == "ConvexHullExpectation("
            "convex_hull_property=region_measure_ch, "
            "expected_variance=None)"
        )
        assert cpe.results is None
        assert cpe.meta.comment == "this is an example"

    def test_empty_locdata(self, caplog):
        cpe = ConvexHullExpectation().compute(LocData())
        assert cpe.results is None
        cpe.plot()
        cpe.hist()
        # plt.show()

        assert caplog.record_tuples == [
            ("locan.analysis.convex_hull_expectation", 30, "Locdata is empty."),
        ]

        plt.close("all")

    def test_compute(self, locdata_2d):
        locdata_2d = deepcopy(locdata_2d)
        collection = LocData.from_collection(locdatas=[locdata_2d, locdata_2d])
        cpe = ConvexHullExpectation().compute(locdata=collection)
        # print(cpe.results.values)
        # print(cpe.results.grouped)
        assert isinstance(cpe.results, ConvexHullExpectationResults)
        assert cpe.results.values.index.tolist() == collection.data.index.tolist()

        assert cpe.results.values["region_measure_ch"].tolist() == pytest.approx(
            [14, 14], rel=0.1
        )

        assert cpe.results.values.columns.tolist() == [
            "localization_count",
            "region_measure_ch",
            "expectation",
            "value_to_expectation_ratio",
        ]
        assert cpe.results.grouped.index.tolist() == [6]
        assert cpe.results.grouped.columns.tolist() == [
            "region_measure_ch_mean",
            "region_measure_ch_std",
            "expectation",
            "expectation_std_pos",
            "expectation_std_neg",
        ]
        assert cpe.results.grouped.expectation.isna().all()

        cpe.plot()
        # plt.show()
        cpe.hist()
        # plt.show()

        plt.close("all")

    def test_compute_with_expectation(self, locdata_2d):
        locdata_2d = deepcopy(locdata_2d)
        collection = LocData.from_collection(locdatas=[locdata_2d, locdata_2d])
        cpe = ConvexHullExpectation(
            convex_hull_property="region_measure_ch", expected_variance=100
        ).compute(locdata=collection)
        # print(cpe.results.values.columns)
        # print(cpe.results.grouped)
        assert cpe.results.values.index.tolist() == collection.data.index.tolist()
        assert cpe.results.values.columns.tolist() == [
            "localization_count",
            "region_measure_ch",
            "expectation",
            "value_to_expectation_ratio",
        ]
        assert cpe.results.grouped.index.tolist() == [6]
        assert cpe.results.grouped.columns.tolist() == [
            "region_measure_ch_mean",
            "region_measure_ch_std",
            "expectation",
            "expectation_std_pos",
            "expectation_std_neg",
        ]

        assert cpe.results.grouped.expectation.tolist() == pytest.approx(
            [321.0], rel=0.1
        )
        assert cpe.results.grouped.expectation_std_pos.tolist() == pytest.approx(
            [198.0], rel=0.1
        )
        assert cpe.results.grouped.expectation_std_neg.tolist() == pytest.approx(
            [132], rel=0.1
        )
        assert cpe.results.values.value_to_expectation_ratio.tolist() == pytest.approx(
            [0.0436137, 0.0436137], rel=1e-3
        )

        cpe.plot()
        # plt.show()
        cpe.hist()
        # plt.show()

        plt.close("all")

    def test_compute_with_expectation_unsorted_index(self, locdata_2d):
        locdata = deepcopy(locdata_2d)
        other_locdata = deepcopy(locdata_2d)
        other_locdata.dataframe.loc[:, "position_x"] = (
            other_locdata.dataframe.loc[:, "position_x"] * 2
        )
        other_locdata.update(dataframe=other_locdata.dataframe.iloc[1:])
        collection = LocData.from_collection(locdatas=[locdata, other_locdata])
        collection.data.index = [3, 1]
        cpe = ConvexHullExpectation(
            convex_hull_property="region_measure_ch", expected_variance=100
        ).compute(locdata=collection)
        assert "region_measure_ch" in locdata.properties.keys()
        assert "region_measure_ch" in collection.data.columns
        assert (
            cpe.results.values.sort_index().index.tolist()
            == collection.data.sort_index().index.tolist()
        )
        assert cpe.results.values.columns.tolist() == [
            "localization_count",
            "region_measure_ch",
            "expectation",
            "value_to_expectation_ratio",
        ]
        assert cpe.results.grouped.index.tolist() == [5, 6]
        assert cpe.results.grouped.columns.tolist() == [
            "region_measure_ch_mean",
            "region_measure_ch_std",
            "expectation",
            "expectation_std_pos",
            "expectation_std_neg",
        ]

        assert cpe.results.grouped.expectation.tolist() == pytest.approx(
            [250.99999999999997, 321.0], rel=0.1
        )
        assert cpe.results.grouped.expectation_std_pos.tolist() == pytest.approx(
            [183.0, 198.0], rel=0.1
        )
        assert cpe.results.grouped.expectation_std_neg.tolist() == pytest.approx(
            [113.99999999999999, 132.0], rel=0.1
        )
        assert (
            cpe.results.values.sort_index().value_to_expectation_ratio.tolist()
            == pytest.approx([0.0756972111553785, 0.04361370716510903], rel=1e-3)
        )

        cpe.plot()
        # plt.show()
        cpe.hist()
        # plt.show()

        plt.close("all")

    def test_compute_with_expectation_subregion(self, locdata_2d):
        locdata_2d = deepcopy(locdata_2d)
        collection = LocData.from_collection(locdatas=[locdata_2d, locdata_2d])
        cpe = ConvexHullExpectation(
            convex_hull_property="subregion_measure_ch", expected_variance=100
        ).compute(locdata=collection)
        # print(cpe.results.values.)
        # print(cpe.results.grouped)
        assert cpe.results.values.index.tolist() == collection.data.index.tolist()
        assert cpe.results.values.columns.tolist() == [
            "localization_count",
            "subregion_measure_ch",
            "expectation",
            "value_to_expectation_ratio",
        ]
        assert cpe.results.grouped.index.tolist() == [6]
        assert cpe.results.grouped.columns.tolist() == [
            "subregion_measure_ch_mean",
            "subregion_measure_ch_std",
            "expectation",
            "expectation_std_pos",
            "expectation_std_neg",
        ]

        cpe.plot()
        # plt.show()
        cpe.hist()
        # plt.show()

        plt.close("all")

    def test_compute_with_expectation_region_measure_3d(self, locdata_3d):
        locdata_3d = deepcopy(locdata_3d)
        collection = LocData.from_collection(locdatas=[locdata_3d, locdata_3d])
        cpe = ConvexHullExpectation(
            convex_hull_property="region_measure_ch", expected_variance=100
        ).compute(locdata=collection)
        # print(cpe.results.values)
        # print(cpe.results.grouped)
        assert cpe.results.values.index.tolist() == collection.data.index.tolist()
        assert cpe.results.values.columns.tolist() == [
            "localization_count",
            "region_measure_ch",
            "expectation",
            "value_to_expectation_ratio",
        ]
        assert cpe.results.grouped.index.tolist() == [6]
        assert cpe.results.grouped.columns.tolist() == [
            "region_measure_ch_mean",
            "region_measure_ch_std",
            "expectation",
            "expectation_std_pos",
            "expectation_std_neg",
        ]

        cpe.plot()
        # plt.show()
        cpe.hist()
        # plt.show()

        plt.close("all")

    def test_compute_without_references(self):
        collection = LocData.from_dataframe(
            dataframe=pd.DataFrame.from_dict(
                {
                    "localization_count": np.array([6, 5, 300]),
                    "position_x": np.array([2.666667, 2.666667, 5]),
                    "position_y": np.array([3.666667, 3.666667, 5]),
                    "region_measure_ch": np.array([14, 14, 100]),
                }
            )
        )
        collection.data.index = [3, 1, 4]
        cpe = ConvexHullExpectation(
            convex_hull_property="region_measure_ch", expected_variance=100
        ).compute(locdata=collection)
        # print(cpe.results.values.columns)
        # print(cpe.results.values[["localization_count", "expectation"]])
        # print(cpe.results.grouped)
        assert "region_measure_ch" in collection.data.columns
        assert (
            cpe.results.values.sort_index().index.tolist()
            == collection.data.sort_index().index.tolist()
        )
        assert cpe.results.values.columns.tolist() == [
            "localization_count",
            "region_measure_ch",
            "expectation",
            "value_to_expectation_ratio",
        ]
        assert cpe.results.grouped.index.tolist() == [5, 6, 300]
        assert cpe.results.grouped.columns.tolist() == [
            "region_measure_ch_mean",
            "region_measure_ch_std",
            "expectation",
            "expectation_std_pos",
            "expectation_std_neg",
        ]

        assert cpe.results.grouped.expectation.tolist() == pytest.approx(
            [250.99999999999997, 321.0, np.nan], rel=0.1, nan_ok=True
        )
        assert cpe.results.grouped.expectation_std_pos.tolist() == pytest.approx(
            [183.0, 198.0, np.nan], rel=0.1, nan_ok=True
        )
        assert cpe.results.grouped.expectation_std_neg.tolist() == pytest.approx(
            [113.99999999999999, 132.0, np.nan], rel=0.1, nan_ok=True
        )
        assert (
            cpe.results.values.sort_index().value_to_expectation_ratio.tolist()
            == pytest.approx(
                [0.05577689243027889, 0.04361370716510903, np.nan],
                rel=1e-3,
                nan_ok=True,
            )
        )

        cpe.plot()
        # plt.show()
        cpe.hist()
        # plt.show()

        plt.close("all")


@pytest.mark.visual
class TestConvexHullPropertyExpectationVisual:
    def test_compute_for_visualization(self, locdata_blobs_2d):
        locdata_blobs_2d = deepcopy(locdata_blobs_2d)
        grouped = locdata_blobs_2d.data.groupby("cluster_label")
        collection = LocData.from_collection(
            locdatas=[
                LocData.from_dataframe(dat.iloc[: (10 - i)])
                for i, (name, dat) in enumerate(grouped)
            ]
        )
        cpe = ConvexHullExpectation(expected_variance=400).compute(locdata=collection)

        print(cpe.results.values.columns)
        print(cpe.results.grouped)

        cpe.plot()
        plt.show()

        cpe.hist()
        plt.show()
        cpe.hist(bin_size=(1, 1000), bin_range=((3, 100), (1, 10_000)), log=False)
        plt.show()
        cpe.hist(bin_size=(1, 1000), bin_range=((3, 100), (1, 10_000)))
        plt.show()

        axes = bh.axis.AxesTuple(
            (
                bh.axis.Regular(20, 3, 503, transform=bh.axis.transform.log),
                bh.axis.Regular(50, 1, 15_000, transform=bh.axis.transform.log),
            )
        )
        cpe.hist(bins=axes)
        plt.show()

        plt.close("all")


class TestConvexHullPropertyExpectationBatch:
    def test_init(self, locdata_2d):
        cpeb = ConvexHullExpectationBatch(meta={"comment": "this is an example"})
        assert (
            str(cpeb) == "ConvexHullExpectationBatch("
            "convex_hull_property=region_measure_ch, "
            "expected_variance=None)"
        )
        assert cpeb.results is None
        assert cpeb.batch is None
        assert cpeb.meta.comment == "this is an example"

    def test_compute_empty_locdata(self, caplog):
        cpeb = ConvexHullExpectationBatch().compute(locdatas=[LocData(), LocData()])
        assert cpeb.results is None
        assert cpeb.batch is None
        assert caplog.record_tuples == [
            ("locan.analysis.convex_hull_expectation", 30, "Locdata is empty."),
            ("locan.analysis.convex_hull_expectation", 30, "Locdata is empty."),
            ("locan.analysis.convex_hull_expectation", 30, "The batch is empty."),
        ]
        cpeb.plot()
        cpeb.hist()
        # plt.show()
        plt.close("all")

    def test_from_batch_empty_item(self, caplog):
        cpeb = ConvexHullExpectationBatch().from_batch(
            batch=[ConvexHullExpectation(), ConvexHullExpectation()]
        )
        assert cpeb.results is None
        assert cpeb.batch is None
        cpeb.plot()
        cpeb.hist()
        # plt.show()

        assert caplog.record_tuples == [
            ("locan.analysis.convex_hull_expectation", 30, "The batch is empty."),
        ]

        plt.close("all")

    def test_compute_empty_locdatas(self, caplog):
        cpeb = ConvexHullExpectationBatch().compute(locdatas=[])
        assert cpeb.results is None
        assert cpeb.batch is None
        assert caplog.record_tuples == [
            ("locan.analysis.convex_hull_expectation", 30, "The batch is empty."),
        ]
        cpeb.plot()
        cpeb.hist()
        # plt.show()
        plt.close("all")

    def test_from_batch_empty(self, caplog):
        cpeb = ConvexHullExpectationBatch().from_batch(batch=[])
        assert cpeb.results is None
        assert cpeb.batch is None
        assert caplog.record_tuples == [
            ("locan.analysis.convex_hull_expectation", 30, "The batch is empty."),
        ]
        cpeb.plot()
        cpeb.hist()
        # plt.show()
        plt.close("all")

    def test_compute(self, locdata_2d):
        collection = LocData.from_dataframe(
            dataframe=pd.DataFrame.from_dict(
                {
                    "localization_count": np.array([6, 5]),
                    "position_x": np.array([2.666667, 2.666667]),
                    "position_y": np.array([3.666667, 3.666667]),
                    "region_measure_ch": np.array([14, 14]),
                }
            )
        )
        collection.data.index = [3, 1]

        collection_2 = LocData.from_dataframe(
            dataframe=pd.DataFrame.from_dict(
                {
                    "localization_count": np.array([6, 5, 300]),
                    "position_x": np.array([2.666667, 2.666667, 5]),
                    "position_y": np.array([3.666667, 3.666667, 5]),
                    "region_measure_ch": np.array([14, 14, 100]),
                }
            )
        )
        collection_2.data.index = [3, 1, 4]

        cpeb = ConvexHullExpectationBatch().compute(locdatas=[collection, collection_2])

        # print(cpeb.results.values)
        # print(cpeb.results.grouped)
        assert isinstance(cpeb.results, ConvexHullExpectationResults)
        assert cpeb.results.values.index.tolist() == list(range(5))

        assert cpeb.results.values["region_measure_ch"].tolist() == pytest.approx(
            [14, 14, 14, 14, 100], rel=0.1
        )

        assert cpeb.results.values.columns.tolist() == [
            "localization_count",
            "region_measure_ch",
            "expectation",
            "value_to_expectation_ratio",
        ]
        assert cpeb.results.grouped.index.tolist() == [5, 6, 300]
        assert cpeb.results.grouped.columns.tolist() == [
            "region_measure_ch_mean",
            "region_measure_ch_std",
            "expectation",
            "expectation_std_pos",
            "expectation_std_neg",
        ]
        assert cpeb.results.grouped.expectation.isna().all()

        cpeb.plot()
        # plt.show()
        cpeb.hist()
        # plt.show()

        plt.close("all")



"""

Transform localization data.

This module takes localization data and applies transformation procedures on
coordinates or other properties.

"""

from __future__ import annotations

import logging
import sys
from collections.abc import Iterable, Sequence
from typing import Any, Literal

import numpy as np
import numpy.typing as npt
import pandas as pd

from locan.data.locdata import LocData
from locan.data.metadata_utils import _modify_meta
from locan.data.validation import _check_loc_properties
from locan.dependencies import HAS_DEPENDENCY, needs_package

if HAS_DEPENDENCY["open3d"]:
    import open3d as o3d


__all__: list[str] = ["transform_affine", "standardize", "overlay"]

logger = logging.getLogger(__name__)


def _transform_affine_numpy(
    points: npt.ArrayLike,
    matrix: npt.ArrayLike | None = None,
    offset: npt.ArrayLike | None = None,
    pre_translation: npt.ArrayLike | None = None,
) -> npt.NDArray[np.float64]:
    """
    Transform `points` by an affine transformation using standard numpy
    procedures.

    Parameters
    ----------
    points
        Points on which to perform the manipulation.
    matrix
        Transformation matrix. If None the unit matrix is used.
        Array with shape (ndim, ndim).
        If None the unit matrix is used.
    offset
        Translation vector.
        Array with shape (ndim,).
        If None a vector of zeros is used.
    pre_translation
        Translation vector for coordinates applied before affine
        transformation. Array with shape (ndim,).
        The reverse translation is applied after the affine transformation.

    Returns
    -------
    npt.NDArray[np.float64]
        Transformed coordinates.
    """
    points_ = np.asarray(points)
    dimension = np.shape(points_)[1]

    matrix_ = np.identity(dimension) if matrix is None else np.asarray(matrix)
    offset_ = np.zeros(dimension) if offset is None else np.asarray(offset)

    if pre_translation is None:
        # transformed_points = np.array(
        # [np.dot(matrix_, point) + offset_ for point in points_]
        # )
        # same function but better performance:
        transformed_points: npt.NDArray[np.float64] = (
            np.einsum("ij, nj -> ni", matrix_, points_) + offset_
        )
    else:
        pre_translation = np.asarray(pre_translation)
        transformed_points = points_ + pre_translation
        # transformed_points = np.array(
        # [np.dot(matrix_, point) + offset_ for point in transformed_points]
        # )
        transformed_points = (
            np.einsum("ij, nj -> ni", matrix_, transformed_points) + offset_
        )
        transformed_points = transformed_points - pre_translation

    return transformed_points


def _homogeneous_matrix(
    matrix: npt.ArrayLike | None = None, offset: npt.ArrayLike | None = None
) -> npt.NDArray[np.float64]:
    """
    Combine transformation matrix and translation vector for dimension d into
    homogeneous (d+1, d+1) transformation
    matrix to be used with homogeneous coordinates.

    Parameters
    ----------
    matrix
        Transformation matrix. If None the unit matrix is used.
        Array with shape (ndim, ndim).
    offset
        Translation vector.
        Array with shape (ndim,).

    Returns
    -------
    npt.NDArray[np.float64]
        Homogeneous transformation matrix to be used with homogeneous
        coordinate vector. Array with shape (ndim+1, ndim+1).
    """
    matrix = np.identity(3) if matrix is None else np.asarray(matrix)
    offset = np.zeros(3) if offset is None else np.asarray(offset)
    dimension = np.shape(matrix)[0]

    if dimension != np.shape(matrix)[1]:
        raise ValueError("The matrix has to be of shape (d, d).")

    if dimension != np.shape(offset)[0]:
        raise ValueError("The matrix and offset must have the same dimension.")

    matrix_ = np.identity(dimension + 1)
    matrix_[0:dimension, 0:dimension] = matrix
    matrix_[:dimension, dimension] = offset

    return matrix_


@needs_package("open3d")
def _transform_affine_open3d(
    points: npt.ArrayLike,
    matrix: npt.ArrayLike | None = None,
    offset: npt.ArrayLike | None = None,
    pre_translation: npt.ArrayLike | None = None,
) -> npt.NDArray[np.float64]:
    """
    Transform `points` or coordinates in `locdata` by an affine
    transformation using open3d.

    Parameters
    ----------
    points
        Points on which to perform the manipulation.
    matrix
        Transformation matrix. If None the unit matrix is used.
        Array with shape (ndim, ndim).
        If None the unit matrix is used.
    offset
        Translation vector.
        Array with shape (ndim,).
        If None a vector of zeros is used.
    pre_translation
        Translation vector for coordinates applied before affine
        transformation. Array with shape (ndim,).
        The reverse translation is applied after the affine transformation.

    Returns
    -------
    npt.NDArray[np.float64]
        Transformed coordinates.
    """
    points_ = np.asarray(points)
    dimension = np.shape(points_)[1]

    matrix_ = np.identity(dimension) if matrix is None else np.asarray(matrix)
    offset_ = np.zeros(dimension) if offset is None else np.asarray(offset)

    # prepare 3d
    if dimension == 2:
        points_3d = np.concatenate([points_, np.zeros((len(points_), 1))], axis=1)
    elif dimension == 3:
        points_3d = points_
    else:
        raise ValueError("Point array has the wrong shape.")

    # points in open3d
    point_cloud = o3d.geometry.PointCloud()
    point_cloud.points = o3d.utility.Vector3dVector(points_3d)

    # transform
    matrix_3d = np.identity(3)
    matrix_3d[:dimension, :dimension] = matrix_
    offset_3d = np.zeros(3)
    offset_3d[:dimension] = offset_
    matrix_homogeneous = _homogeneous_matrix(matrix_3d, offset_3d)

    if pre_translation is None:
        point_cloud.transform(matrix_homogeneous)
    else:
        pre_translation_3d = np.zeros(3)
        pre_translation_3d[:dimension] = pre_translation
        point_cloud.translate(pre_translation_3d)
        point_cloud.transform(matrix_homogeneous)
        point_cloud.translate(-pre_translation_3d)

    # return correct dimension
    if dimension == 2:
        transformed_points = np.asarray(point_cloud.points)[:, 0:2]
    else:  # if dimension == 3:
        transformed_points = np.asarray(point_cloud.points)

    return transformed_points


def transform_affine(
    locdata: npt.ArrayLike | LocData,
    matrix: npt.ArrayLike | None = None,
    offset: npt.ArrayLike | None = None,
    pre_translation: npt.ArrayLike | None = None,
    method: Literal["numpy", "open3d"] = "numpy",
) -> npt.NDArray[np.float64] | LocData:
    """
    Transform `points` or coordinates in `locdata` by an affine transformation.

    Parameters
    ----------
    locdata
        Localization data on which to perform the manipulation.
    matrix
        Transformation matrix. If None the unit matrix is used.
        Array with shape (ndim, ndim).
        If None the unit matrix is used.
    offset
        Translation vector.
        Array with shape (ndim,).
        If None a vector of zeros is used.
    pre_translation
        Translation vector for coordinates applied before affine
        transformation. Array with shape (ndim,).
        The reverse translation is applied after the affine transformation.
    method
        The method (i.e. library or algorithm) used for computation.
        One of 'numpy', 'open3d'.

    Returns
    -------
    npt.NDArray[np.float64] | LocData
        New localization data with transformed coordinates.
    """
    local_parameter = locals()

    if len(locdata) == 0:  # type: ignore
        return locdata  # type: ignore

    # adjust input
    if isinstance(locdata, LocData):
        points = locdata.coordinates
    else:
        points = np.asarray(locdata)

    if method == "numpy":
        transformed_points = _transform_affine_numpy(
            points, matrix=matrix, offset=offset, pre_translation=pre_translation
        )
    elif method == "open3d":
        transformed_points = _transform_affine_open3d(
            points, matrix=matrix, offset=offset, pre_translation=pre_translation
        )
    else:
        raise ValueError(f"Method {method} is not available.")

    # prepare output
    if isinstance(locdata, LocData):
        # new LocData object
        new_dataframe = locdata.data.copy()
        df = pd.DataFrame(
            transformed_points,
            columns=locdata.coordinate_keys,
            index=locdata.data.index,
        )
        # cast dtypes
        columns_intersection = list(set(df.columns) & set(locdata.coordinate_keys))
        df = df.astype(new_dataframe[columns_intersection].dtypes)

        new_dataframe.update(df)
        new_locdata = LocData.from_dataframe(new_dataframe)

        # update metadata
        meta_ = _modify_meta(
            locdata,
            new_locdata,
            function_name=sys._getframe().f_code.co_name,
            parameter=local_parameter,
            meta=None,
        )
        new_locdata.meta = meta_

        return new_locdata

    else:
        return transformed_points


def standardize(
    locdata: LocData,
    loc_properties: list[str] | None = None,
    with_mean: bool = True,
    with_std: bool = True,
) -> LocData:
    """
    Transform locdata properties by centering to the mean
    and property-wise scaling to unit standard deviation (variance).

    Note
    -----
    This function makes use of :func:sklearn.preprocessing.scale
    and thus works with a biased estimator for the standard deviation.

    Parameters
    ----------
    locdata
        Localization data to be standardized.
    loc_properties
        Localization properties to be standardized.
        If None The coordinate_values of locdata are used.
    with_mean
        If True center to the mean.
    with_std
        If True scale to unit standard deviation (variance).

    Returns
    -------
    LocData
        New localization data with standardized properties.
    """
    local_parameter = locals()

    if len(locdata) == 0:
        return locdata

    labels_ = _check_loc_properties(locdata, loc_properties)
    data = locdata.data[labels_].values

    from sklearn.preprocessing import scale

    transformed_data = scale(data, with_mean=with_mean, with_std=with_std)

    new_dataframe = locdata.data.copy()
    df = pd.DataFrame(
        transformed_data,
        columns=labels_,
        index=locdata.data.index,
    )
    # cast dtypes
    columns_intersection = list(set(df.columns) & set(labels_))
    df = df.astype(new_dataframe[columns_intersection].dtypes)
    new_dataframe.update(df)

    new_locdata = LocData.from_dataframe(new_dataframe)

    # update metadata
    meta_ = _modify_meta(
        locdata,
        new_locdata,
        function_name=sys._getframe().f_code.co_name,
        parameter=local_parameter,
        meta=None,
    )
    new_locdata.meta = meta_

    return new_locdata


def overlay(
    locdatas: Iterable[LocData],
    centers: Iterable[Any] | str | None = "centroid",
    orientations: Sequence[int | float | str | None] | str | None = None,
) -> LocData:
    """
    Translate locdatas to their common center and rotate according to their
    orientation.

    Parameters
    ----------
    locdatas
        Localization data to overlay.
    centers
        centers to which locdatas are translated.
        Must have the same length as locdatas.
        One of `centroid`, `ch`, 'bb', 'obb', or 'region'.
        If None, no translation is applied.
    orientations
        Orientation value to use in degree.
        Must have the same length as locdatas.
        If str, it must be one of `orientation_im`, `orientation_obb`.
        If None, no rotation is applied.

    Returns
    -------
    LocData
        Collection with transformed locdatas.
    """
    local_parameter = locals()

    if not locdatas or not isinstance(locdatas, (tuple, list)):
        raise TypeError("locdatas must be a list of LocData objects.")

    dimensions = {locdata.dimension for locdata in locdatas}
    if len(dimensions) == 1:
        dimension = dimensions.pop()
    else:
        raise ValueError(
            "Some elements of `locdatas` have a different dimension. "
            "All locdatas must have the same dimension."
        )

    if centers is None or isinstance(centers, str):
        centers = [centers] * len(locdatas)
    else:
        if len(centers) != len(locdatas):  # type: ignore
            raise ValueError("Centers must have the same length as locdatas.")

    if orientations is None or isinstance(orientations, str):
        orientations = [orientations] * len(locdatas)
    else:
        if len(orientations) != len(locdatas):
            raise ValueError("orientations must have the same length as locdatas.")

    transformed_locdatas = []
    for locdata, centre, orientation in zip(locdatas, centers, orientations):
        # translation
        if centre is None:
            transformed_locdata = locdata
        elif not isinstance(centre, str):
            transformed_locdata = transform_affine(
                locdata, matrix=None, offset=np.multiply(centre, -1)
            )
        elif centre == "centroid":
            transformed_locdata = transform_affine(
                locdata, matrix=None, offset=np.multiply(locdata.centroid, -1)
            )
        elif centre == "bb":
            transformed_locdata = transform_affine(
                locdata,
                matrix=None,
                offset=np.multiply(locdata.bounding_box.region.centroid, -1),
            )
        elif centre == "obb":
            transformed_locdata = transform_affine(
                locdata,
                matrix=None,
                offset=np.multiply(locdata.oriented_bounding_box.region.centroid, -1),
            )
        elif centre == "ch":
            transformed_locdata = transform_affine(
                locdata,
                matrix=None,
                offset=np.multiply(locdata.convex_hull.region.centroid, -1),
            )
        elif centre == "region":
            transformed_locdata = transform_affine(
                locdata, matrix=None, offset=np.multiply(locdata.region.centroid, -1)
            )
        else:
            raise ValueError(f"Value for centre={centre} is not defined.")

        # prepare rotation
        if orientation is None:
            matrix = np.identity(dimension)
        elif not isinstance(orientation, str):
            theta = -np.radians(orientation)
            cos_, sin_ = np.cos(theta), np.sin(theta)
            matrix = np.array(((cos_, -sin_), (sin_, cos_)))
        else:
            if dimension != 2:
                raise NotImplementedError(
                    "Rotation has only been implemented for 2 dimensions."
                )
            if orientation == "orientation_obb":
                locdata.oriented_bounding_box  # update oriented_bounding_box  # noqa B018
            elif orientation == "orientation_im":
                locdata.inertia_moments  # update inertia_moments  # noqa B018
            else:
                raise ValueError(f"orientation={orientation} is undefined.")

            theta = -np.radians(locdata.properties[orientation])
            cos_, sin_ = np.cos(theta), np.sin(theta)
            matrix = np.array(((cos_, -sin_), (sin_, cos_)))

        # rotation
        transformed_locdata = transform_affine(
            transformed_locdata, matrix=matrix, offset=None
        )

        transformed_locdatas.append(transformed_locdata)

    new_locdata = LocData.from_collection(transformed_locdatas)

    # update metadata
    del new_locdata.meta.history[:]
    new_locdata.meta.history.add(
        name=sys._getframe().f_code.co_name, parameter=str(local_parameter)
    )

    return new_locdata



"""
Compute subpixel bias in localization data.

Subpixel bias in localization coordinates may arise depending on the
localization algorithm [1]_.

References
----------
.. [1] Gould, T. J., Verkhusha, V. V. & Hess, S. T.,
   Imaging biological structures with fluorescence photoactivation
   localization microscopy. Nat. Protoc. 4 (2009), 291–308.
"""

from __future__ import annotations

import logging
import sys
from collections.abc import Sequence
from typing import TYPE_CHECKING, Any, cast

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

if TYPE_CHECKING:
    import matplotlib as mpl

    from locan.data.locdata import LocData

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd

from locan.analysis import metadata_analysis_pb2
from locan.analysis.analysis_base import _Analysis

__all__: list[str] = ["SubpixelBias"]

logger = logging.getLogger(__name__)


# The algorithms


def _subpixel_bias(
    locdata: LocData, pixel_size: int | float | Sequence[int | float]
) -> pd.DataFrame:
    coordinate_labels = locdata.coordinate_keys
    coordinates = locdata.coordinates.T

    if np.ndim(pixel_size) == 0:
        pixel_size = cast("int | float", pixel_size)
        pixel_sizes: Sequence[int | float] = [pixel_size] * len(coordinate_labels)
    else:
        pixel_size = cast("Sequence[int | float]", pixel_size)
        if len(pixel_size) != len(coordinate_labels):
            raise TypeError("There must be given a pixel_size for each coordinate.")
        else:
            pixel_sizes = pixel_size

    coordinates_modulo = [
        np.remainder(coordinates_, pixel_size_)
        for coordinates_, pixel_size_ in zip(coordinates, pixel_sizes)
    ]

    results = {
        label_ + "_modulo": values_
        for label_, values_ in zip(coordinate_labels, coordinates_modulo)
    }

    return pd.DataFrame(results)


# The specific analysis classes


class SubpixelBias(_Analysis):
    """
    Check for subpixel bias by computing the modulo of localization coordinates
    for each localization's spatial coordinate in locdata.

    Parameters
    ----------
    meta : locan.analysis.metadata_analysis_pb2.AMetadata | None
        Metadata about the current analysis routine.
    pixel_size : int | float | Sequence[int | float]
        Camera pixel size in coordinate units.

    Attributes
    ----------
    count : int
        A counter for counting instantiations.
    parameter : dict
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    results : pandas.DataFrame
        The number of localizations per frame or
        the number of localizations per frame normalized to region_measure(hull).
    """

    count = 0

    def __init__(
        self,
        meta: metadata_analysis_pb2.AMetadata | None = None,
        pixel_size: int | float | Sequence[int | float] | None = None,
    ) -> None:
        parameters = self._get_parameters(locals())
        super().__init__(**parameters)
        self.results = None

    def compute(self, locdata: LocData) -> Self:
        """
        Run the computation.

        Parameters
        ----------
        locdata
            Localization data.

        Returns
        -------
        Self
        """
        if not len(locdata):
            logger.warning("Locdata is empty.")
            return self

        self.results = _subpixel_bias(locdata=locdata, **self.parameter)
        return self

    def hist(
        self,
        ax: mpl.axes.Axes | None = None,
        bins: int | Sequence[int | float] | str = "auto",
        log: bool = True,
        **kwargs: Any,
    ) -> mpl.axes.Axes:
        """
        Provide histogram as :class:`matplotlib.axes.Axes` object showing
        hist(results). Nan entries are ignored.

        Parameters
        ----------
        ax
            The axes on which to show the image
        bins
            Bin specifications (passed to :func:`matplotlib.hist`).
        log
            Flag for plotting on a log scale.
        kwargs
            Other parameters passed to :func:`matplotlib.pyplot.hist`.

        Returns
        -------
        matplotlib.axes.Axes
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if self.results is None:
            return ax

        ax.hist(
            self.results.dropna(axis=0).to_numpy(),
            bins=bins,
            **dict(dict(density=True, log=log), **kwargs),
            histtype="step",
            label=self.results.columns,
        )
        ax.set(
            title="Subpixel Bias",
            xlabel="position_modulo_pixel_size",
            ylabel="PDF",
        )

        return ax



"""
Module to deal with required and optional dependencies.

Optional dependencies are defined in pyproject.toml.

In any module that requires an optional dependency the import should be
conditioned:

    if HAS_DEPENDENCY["package"]: import package

Any function that makes use of the optional dependency should be decorated with

    @needs_package("package")


CONSTANTS

.. autosummary::
   :toctree: ./

   INSTALL_REQUIRES
   EXTRAS_REQUIRE
   IMPORT_NAMES
   HAS_DEPENDENCY
"""

from __future__ import annotations

import importlib.metadata
import importlib.util
import logging
import os
import re
import sys
from collections.abc import Callable, Iterable
from enum import Enum
from functools import wraps
from typing import Any, TypeVar

if sys.version_info >= (3, 10):
    from typing import ParamSpec
else:
    from typing_extensions import ParamSpec

logger = logging.getLogger(__name__)

__all__: list[str] = [
    "needs_package_version",
    "needs_package",
    "IMPORT_NAMES",
    "INSTALL_REQUIRES",
    "EXTRAS_REQUIRE",
    "HAS_DEPENDENCY",
    "QtBindings",
]

F = TypeVar("F", bound=Callable[..., Any])
P = ParamSpec("P")
T = TypeVar("T")


def _has_dependency_factory(
    packages: Iterable[str], import_names: dict[str, str] | None = None
) -> dict[str, bool]:
    if import_names is None:
        import_names = IMPORT_NAMES
    has_dependency = dict()
    for package in packages:
        key = import_names.get(package, package)
        value = importlib.util.find_spec(key) is not None
        has_dependency[key] = value
    return has_dependency


def needs_package(
    package: str,
    import_names: dict[str, str] | None = None,
    has_dependency: dict[str, bool] | None = None,
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """
    Function that returns a decorator to check for optional dependency.

    Parameters
    ----------
    package
        Package or dependency name that needs to be imported.
    import_names
        Mapping of package names onto import names.
    has_dependency
        Dictionary with bool indicator if package (import name) is available.

    Returns
    -------
    callable
        A decorator that raises ImportError if package is not available.
    """
    if import_names is None:
        import_names = IMPORT_NAMES
    if has_dependency is None:
        has_dependency = HAS_DEPENDENCY
    import_name = import_names.get(package, package)

    def decorator(func: Callable[P, T]) -> Callable[P, T]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            if not has_dependency[import_name]:  # type: ignore
                raise ImportError(
                    f"Function {func} needs {import_name} which cannot be imported."
                )
            return func(*args, **kwargs)

        return wrapper

    return decorator


def _get_dependencies(package: str) -> tuple[set[str], set[str]]:
    """
    Read out all required and optional (extra) dependencies for package.

    Parameters
    ----------
    package
        Package or dependency name to be checked.

    Returns
    -------
    tuple[set[str], set[str]]
        PyPI names
    """
    requires = importlib.metadata.requires(package)
    if requires is None:
        required_dependencies: set[str] = set()
        extra_dependencies: set[str] = set()
    else:
        pattern = r"[\w-]+"
        required_dependencies = set()
        extra_dependencies = set()
        for item in requires:
            match = re.match(pattern, item)
            if match is None:
                pass
            else:
                if "extra ==" in item:
                    extra_dependencies.add(match.group())
                else:
                    required_dependencies.add(match.group())
    return required_dependencies, extra_dependencies


#: List of required dependencies (PyPi package names)
INSTALL_REQUIRES: set[str] = set()

#: List of optional dependencies (PyPi package names)
EXTRAS_REQUIRE: set[str] = set()
INSTALL_REQUIRES, EXTRAS_REQUIRE = _get_dependencies(package="locan")

#: A dictionary mapping PyPi package names to import names if they are different
IMPORT_NAMES = dict()
IMPORT_NAMES["ruamel"] = "ruamel.yaml"
IMPORT_NAMES["fast-histogram"] = "fast_histogram"
IMPORT_NAMES["boost-histogram"] = "boost_histogram"
IMPORT_NAMES["protobuf"] = "google.protobuf"
IMPORT_NAMES["scikit-image"] = "skimage"
IMPORT_NAMES["scikit-learn"] = "sklearn"
IMPORT_NAMES["pytest-qt"] = "pytestqt"
IMPORT_NAMES["mpl-scatter-density"] = "mpl_scatter_density"

#: A dictionary indicating if dependency is available.
HAS_DEPENDENCY = _has_dependency_factory(
    packages=INSTALL_REQUIRES.union(EXTRAS_REQUIRE)
)


# Possible python bindings to interact with QT
class QtBindings(Enum):
    """
    Python bindings to interact with Qt.
    """

    NONE = ""
    PYSIDE2 = "pyside2"
    PYQT5 = "pyqt5"
    PYSIDE6 = "pyside6"
    PYQT6 = "pyqt6"


def _set_qt_binding(qt_binding: QtBindings | str) -> str:
    """
    Check if qtpy can import `qt_binding` and return the qt_binding that will be used.
    Checks os.environ["QT_API"] first.
    If os.environ["QT_API"] is set it will take precedence over `qt_bindings`.

    Note
    -----
    This function must be used before qtpy is imported for the first time
    to be effective.

    Returns
    -------
    str
        qt-bindings or empty string
    """
    if isinstance(qt_binding, QtBindings):
        qt_api = qt_binding.value
    elif qt_binding:
        qt_api = qt_binding.lower()
    else:
        qt_api = ""

    if "qtpy" in sys.modules:
        logger.warning("qtpy has already been loaded. QT_BINDING cannot be changed.")
    elif os.environ.get("QT_API", ""):
        qt_api = os.environ["QT_API"]
    else:
        if qt_api:
            os.environ["QT_API"] = qt_api

    try:
        from qtpy import (  # noqa: F401  # import API alone does not raise ModuleNotFoundError if no module is available
            API,
            QtCore,
        )

        if qt_api and qt_api != API:
            logger.warning(
                f"QT_BINDING {qt_api} is not available - {API} is used instead."
            )
            os.environ["QT_API"] = API
        qt_api = API
    except ImportError:
        if qt_api:
            logger.warning("QT_BINDING is not available.")
        qt_api = ""

    return qt_api


def needs_package_version(
    package: str,
    major: str,
) -> Callable[[Callable[P, T]], Callable[P, T]]:
    """
    Function that returns a decorator to check for package.version starting with major.

    Parameters
    ----------
    package
        Package or dependency name to be checked.
    major
        Major version number that is required.

    Returns
    -------
    callable
        A decorator that raises RuntimeError if package is not available.
    """

    def decorator(func: Callable[P, T]) -> Callable[P, T]:
        @wraps(func)
        def wrapper(*args: P.args, **kwargs: P.kwargs) -> T:
            version = importlib.metadata.version(package)
            if not version.startswith(major):  # type: ignore
                raise RuntimeError(
                    f"Function {func} needs {package}~{major}. "
                    f"The installed version is {package}={version}."
                )
            return func(*args, **kwargs)

        return wrapper

    return decorator



"""
Analyze the distribution of a localization property in 2D.

Analyze the distribution of a localization property as function of two other
localization properties in 2D.
E.g. looking at how the local background is distributed over localization
coordinates helps to characterize the
illumination profile in SMLM experiments.
"""

from __future__ import annotations

import logging
import sys
from collections.abc import Callable, Sequence
from typing import TYPE_CHECKING, Any, Literal, NamedTuple

if sys.version_info >= (3, 11):
    from typing import Self
else:
    from typing_extensions import Self

if TYPE_CHECKING:
    import boost_histogram as bh
    import matplotlib as mpl

    from locan.data.locdata import LocData

import matplotlib.pyplot as plt
import numpy as np
import numpy.typing as npt
from lmfit import Model, Parameters
from lmfit.model import ModelResult

from locan.analysis import metadata_analysis_pb2
from locan.analysis.analysis_base import _Analysis
from locan.configuration import COLORMAP_DEFAULTS
from locan.process.aggregate import Bins, histogram
from locan.visualize import get_colormap
from locan.visualize.transform import Trafo, adjust_contrast

__all__: list[str] = ["LocalizationProperty2d"]

logger = logging.getLogger(__name__)


# The algorithms


class FitImageResults(NamedTuple):
    image: npt.NDArray[Any]
    bins: Bins
    label: list[str]
    model_result: ModelResult


# model fit function in 2d
def _gauss_2d(
    x: npt.ArrayLike,
    y: npt.ArrayLike,
    amplitude: float,
    center_x: float,
    center_y: float,
    sigma_x: float,
    sigma_y: float,
) -> float | npt.NDArray[np.float64]:
    """
    2D Gauss function of variables (x, y).

    Parameters
    ----------
    x
        all x values
    y
        all y values
    amplitude
        amplitude
    center_x
        x-shift
    center_y
        y-shift
    sigma_x
        standard deviation x
    sigma_y
        standard deviation y

    Returns
    -------
    float | npt.NDArray[np.float64]
    """
    x_ = np.asarray(x)
    y_ = np.asarray(y)
    amplitude, center_x, center_y, sigma_x, sigma_y = (  # type: ignore[assignment]
        np.float64(number)
        for number in (amplitude, center_x, center_y, sigma_x, sigma_y)
    )
    return_value: float | npt.NDArray[np.float64] = amplitude * np.exp(
        -(
            (x_ - center_x) ** 2 / (2.0 * sigma_x**2)
            + (y_ - center_y) ** 2 / (2.0 * sigma_y**2)
        )
    )
    return return_value


def _fit_image(data: npt.ArrayLike, bin_range: npt.ArrayLike) -> ModelResult:
    """
    Fit 2D Gauss function to image data.

    Parameters
    ----------
    data
        arrays with corresponding values for x, y, z
        of shape (3, n_image_values)
    bin_range
        range as returned from histogram().

    Returns
    -------
    lmfit.model.ModelResult
        The fit results.
    """

    # prepare 1D lmfit model from 2D model function
    def model_function(
        points: npt.ArrayLike,
        amplitude: float = 1,
        center_x: float = 0,
        center_y: float = 0,
        sigma_x: float = 1,
        sigma_y: float = 1,
    ) -> npt.NDArray[Any]:
        points = np.asarray(points)
        return np.ravel(
            _gauss_2d(*points.T, amplitude, center_x, center_y, sigma_x, sigma_y)  # type: ignore[call-arg]
        )

    model = Model(model_function, nan_policy="omit")
    # print(model.param_names, model.independent_vars)

    data = np.asarray(data)
    bin_range = np.asarray(bin_range)

    # instantiate lmfit Parameters
    params = Parameters()
    params.add("amplitude", value=np.amax(data[2]))
    centers = np.add(bin_range[:, 0], np.diff(bin_range).flatten() / 2)
    params.add("center_x", value=centers[0])
    params.add("center_y", value=centers[1])
    sigmas = np.diff(bin_range).ravel() / 4
    params.add("sigma_x", value=sigmas[0])
    params.add("sigma_y", value=sigmas[1])

    # fit
    model_result = model.fit(data[2], points=data[0:2].T, params=params)

    return model_result


def _localization_property2d(
    locdata: LocData,
    loc_properties: str | Sequence[str] | None = None,
    other_property: str | None = None,
    bins: Bins | bh.axis.Axis | bh.axis.AxesTuple | None = None,
    n_bins: int | Sequence[int] | None = None,
    bin_size: float | Sequence[float] | Sequence[Sequence[float]] | None = 10,
    bin_edges: Sequence[float] | Sequence[Sequence[float]] | None = None,
    bin_range: (
        tuple[float, float]
        | Sequence[float]
        | Sequence[Sequence[float]]
        | Literal["zero", "link"]
        | None
    ) = None,
    rescale: int | str | Trafo | Callable[..., Any] | bool | None = None,
) -> FitImageResults:
    # bin localization data
    img, bins, label = histogram(
        locdata=locdata,
        loc_properties=loc_properties,
        other_property=other_property,
        bins=bins,
        n_bins=n_bins,
        bin_size=bin_size,
        bin_edges=bin_edges,
        bin_range=bin_range,
    )
    img = adjust_contrast(img, rescale)

    # prepare one-dimensional data
    x, y = bins.bin_centers
    xx, yy = np.meshgrid(x, y, indexing="ij")
    # eliminate image zeros
    positions = np.nonzero(~np.isnan(img))

    data_0 = xx[positions].flatten()
    data_1 = yy[positions].flatten()
    data_2 = img[positions].flatten()

    data = np.stack((data_0, data_1, data_2))

    model_result = _fit_image(data, np.array(bins.bin_range))
    results = FitImageResults(img, bins, label, model_result)
    return results


# The specific analysis classes


class LocalizationProperty2d(_Analysis):
    """
    Analyze 2d distribution of histogram for two localization properties.

    Fit a two dimensional Gauss distribution.

    Parameters
    ----------
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    loc_properties : Sequence[str] | None
        Localization properties to be grouped into bins.
        If None, the coordinate_values of locdata are used.
    other_property : str | None
        Localization property (columns in locdata.data) that is averaged in each pixel.
        If None, the localization counts are shown.
    bins : Bins | boost_histogram.axis.Axis | boost_histogram.axis.AxesTuple | None
        The bin specification as defined in :class:`Bins`
    bin_edges : Sequence[float] | Sequence[Sequence[float]] | None
        Bin edges for all or each dimension
        with shape (dimension, n_bin_edges).
    bin_range : tuple[float, float] | Sequence[float] | Sequence[Sequence[float]] | str | None
        Minimum and maximum edge for all or each dimensions
        with shape (2,) or (dimension, 2).
        If None (min, max) ranges are determined from data and returned;
        if 'zero' (0, max) ranges with max determined from data are returned.
        if 'link' (min_all, max_all) ranges with min and max determined from
        all combined data are returned.
    n_bins : int | Sequence[int] | None
        The number of bins for all or each dimension.
        5 yields 5 bins in all dimensions.
        (2, 5) yields 2 bins for one dimension and 5 for the other dimension.
    bin_size : float | Sequence[float] | Sequence[Sequence[float]] | None
        The size of bins for all or each bin and for all or each dimension
        with shape (dimension,) or (dimension, n_bins).
        5 would describe bin_size of 5 for all bins in all dimensions.
        ((2, 5),) yield bins of size (2, 5) for one dimension.
        (2, 5) yields bins of size 2 for one dimension and 5 for the other
        dimension.
        ((2, 5), (1, 3)) yields bins of size (2, 5) for one dimension and
        (1, 3) for the other dimension.
        To specify arbitrary sequence of `bin_size` use `bin_edges` instead.
    rescale : int | str | locan.Trafo | Callable[..., Any] | bool | None
        Transformation as defined in :class:`locan.Trafo` or by
        transformation function.
        For None or False no rescaling occurs.
        Legacy behavior:
        For tuple with upper and lower bounds provided in percent,
        rescale intensity values to be within percentile of max and min
        intensities.
        For 'equal' intensity values are rescaled by histogram equalization.

    Attributes
    ----------
    count : int
        A counter for counting instantiations (class attribute).
    parameter : dict
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    results : lmfit.model.ModelResult
        Computed fit results.
    """

    def __init__(
        self,
        meta: metadata_analysis_pb2.AMetadata | None = None,
        loc_properties: Sequence[str] | None = None,
        other_property: str | None = "local_background",
        bins: Bins | bh.axis.Axis | bh.axis.AxesTuple | None = None,
        n_bins: int | Sequence[int] | None = None,
        bin_size: float | Sequence[float] | Sequence[Sequence[float]] | None = 100,
        bin_edges: Sequence[float] | Sequence[Sequence[float]] | None = None,
        bin_range: (
            tuple[float, float]
            | Sequence[float]
            | Sequence[Sequence[float]]
            | str
            | None
        ) = None,
        rescale: int | str | Trafo | Callable[..., Any] | bool | None = None,
    ) -> None:
        parameters = self._get_parameters(locals())
        super().__init__(**parameters)
        self.results = None

    def compute(self, locdata: LocData) -> Self:
        """
        Run the computation.

        Parameters
        ----------
        locdata
            Localization data.

        Returns
        -------
        Self
        """
        if not len(locdata):
            logger.warning("Locdata is empty.")
            return self

        self.results = _localization_property2d(locdata=locdata, **self.parameter)
        return self

    def report(self) -> None:
        if self.results is None:
            logger.warning("No results available")
            return

        print("Fit results for:\n")
        print(self.results.model_result.fit_report(min_correl=0.25))
        # print(self.results.fit_results.best_values)

        # judge fit parameter
        max_fit_value = self.results.model_result.best_fit.max()
        min_fit_value = self.results.model_result.best_fit.min()
        ratio = (max_fit_value - min_fit_value) / max_fit_value
        print(f"Maximum fit value in image: {max_fit_value:.3f}")
        print(f"Minimum fit value in image: {min_fit_value:.3f}")
        print(f"Fit value variation over image range: {ratio:.2f}")

    def plot(self, ax: mpl.axes.Axes | None = None, **kwargs: Any) -> mpl.axes.Axes:
        """
        Provide histogram as :class:`matplotlib.axes.Axes` object showing plot(results).

        Parameters
        ----------
        ax
            The axes on which to show the image
        kwargs
            Other parameters passed to :func:`matplotlib.pyplot.contour`.

        Returns
        -------
        matplotlib.axes.Axes
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if self.results is None:
            return ax

        ax.imshow(
            self.results.image.T,
            cmap="viridis",
            origin="lower",
            extent=np.ravel(self.results.bins.bin_range),  # type: ignore
        )

        x, y = self.results.bins.bin_centers
        xx, yy = np.meshgrid(x, y, indexing="ij")
        zz = np.stack((xx, yy), axis=-1).reshape((np.prod(xx.shape), 2))
        z = self.results.model_result.eval(points=zz)

        contourset = ax.contour(
            x,
            y,
            z.reshape((len(x), len(y))).T,
            8,
            **dict(dict(colors="gray"), **kwargs),
        )
        plt.clabel(contourset, fontsize=9, inline=1)
        ax.set(
            title=self.parameter["other_property"],
            xlabel=self.results.label[0],
            ylabel=self.results.label[1],
        )

        return ax

    def plot_residuals(
        self, ax: mpl.axes.Axes | None = None, **kwargs: Any
    ) -> mpl.axes.Axes:
        """
        Provide histogram as :class:`matplotlib.axes.Axes` object showing plot(results).

        Parameters
        ----------
        ax
            The axes on which to show the image
        kwargs
            Other parameters passed to :func:`matplotlib.pyplot.contour`.

        Returns
        -------
        matplotlib.axes.Axes
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if self.results is None:
            return ax

        x, y = self.results.bins.bin_centers
        xx, yy = np.meshgrid(x, y, indexing="ij")
        zz = np.stack((xx, yy), axis=-1).reshape((np.prod(xx.shape), 2))
        z = self.results.model_result.eval(points=zz)

        residuals = np.where(
            self.results.image == 0,
            np.nan,
            z.reshape((len(x), len(y))) - self.results.image,
        )
        max_absolute_value = max([abs(np.nanmin(residuals)), abs(np.nanmax(residuals))])
        ax.imshow(
            residuals.T,
            cmap=get_colormap(colormap=COLORMAP_DEFAULTS["DIVERGING"]).matplotlib,
            origin="lower",
            extent=np.ravel(self.results.bins.bin_range),  # type: ignore
            vmin=(-max_absolute_value),
            vmax=max_absolute_value,
        )

        contourset = ax.contour(
            x,
            y,
            z.reshape((len(x), len(y))).T,
            8,
            **dict(dict(colors="gray"), **kwargs),
        )
        plt.clabel(contourset, fontsize=9, inline=1)

        ax.set(
            title=self.parameter["other_property"],
            xlabel=self.results.label[0],
            ylabel=self.results.label[1],
        )

        return ax

    def plot_deviation_from_mean(
        self, ax: mpl.axes.Axes | None = None
    ) -> mpl.axes.Axes:
        """
        Provide histogram as :class:`matplotlib.axes.Axes` object showing
        plot(results).

        Parameters
        ----------
        ax
            The axes on which to show the image

        Returns
        -------
        matplotlib.axes.Axes
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if self.results is None:
            return ax

        positions = np.nonzero(self.results.image)
        mean_value = np.nanmean(self.results.image[positions])
        deviations = np.where(
            self.results.image == 0, np.nan, self.results.image - mean_value
        )
        max_absolute_value = max(
            [abs(np.nanmin(deviations)), abs(np.nanmax(deviations))]
        )
        ax.imshow(
            deviations.T,
            cmap=get_colormap(colormap=COLORMAP_DEFAULTS["DIVERGING"]).matplotlib,
            origin="lower",
            extent=np.ravel(self.results.bins.bin_range),  # type: ignore
            vmin=(-max_absolute_value),
            vmax=max_absolute_value,
        )

        ax.set(
            title=f"{self.parameter['other_property']} - deviation from mean",
            xlabel=self.results.label[0],
            ylabel=self.results.label[1],
        )

        return ax

    def plot_deviation_from_median(
        self, ax: mpl.axes.Axes | None = None
    ) -> mpl.axes.Axes:
        """
        Provide histogram as :class:`matplotlib.axes.Axes` object showing
        plot(results).

        Parameters
        ----------
        ax
            The axes on which to show the image

        Returns
        -------
        matplotlib.axes.Axes
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if self.results is None:
            return ax

        positions = np.nonzero(self.results.image)
        median_value = np.nanmedian(self.results.image[positions])
        deviations = np.where(
            self.results.image == 0, np.nan, self.results.image - median_value
        )
        max_absolute_value = max(
            [abs(np.nanmin(deviations)), abs(np.nanmax(deviations))]
        )
        ax.imshow(
            deviations.T,
            cmap=get_colormap(colormap=COLORMAP_DEFAULTS["DIVERGING"]).matplotlib,
            origin="lower",
            extent=np.ravel(self.results.bins.bin_range),  # type: ignore
            vmin=(-max_absolute_value),
            vmax=max_absolute_value,
        )

        ax.set(
            title=f"{self.parameter['other_property']} - deviation from median",
            xlabel=self.results.label[0],
            ylabel=self.results.label[1],
        )

        return ax



"""
command-line interface
"""

from __future__ import annotations

import argparse
import sys

from locan import __version__
from locan.scripts.script_check import _add_arguments as _add_arguments_check
from locan.scripts.script_draw_roi import _add_arguments as _add_arguments_draw_roi
from locan.scripts.script_napari import _add_arguments as _add_arguments_napari
from locan.scripts.script_rois import _add_arguments as _add_arguments_rois
from locan.scripts.script_show_versions import (
    _add_arguments as _add_arguments_show_versions,
)
from locan.scripts.script_test import _add_arguments as _add_arguments_test
from locan.utils.system_information import show_versions


def main(args: list[str] | None = None) -> None:
    if args is None:
        args = sys.argv[1:]

    parser = argparse.ArgumentParser(description="Entry point for locan.")

    parser.add_argument(
        "--version",
        action="version",
        version=f"locan version {__version__}",
        help="show version and exit.",
    )

    parser.add_argument(
        "--info", action="store_true", help="show system information and exit."
    )

    subparsers = parser.add_subparsers(dest="command")

    # parser for the command draw_roi_mpl
    parser_draw_roi = subparsers.add_parser(
        name="draw_roi_mpl", description="Set roi by drawing a boundary in mpl."
    )
    _add_arguments_draw_roi(parser_draw_roi)

    # parser for the command rois
    parser_rois = subparsers.add_parser(
        name="rois", description="Define rois by adding shapes in napari."
    )
    _add_arguments_rois(parser_rois)

    # parser for the command check
    parser_check = subparsers.add_parser(
        name="check", description="Show localizations in original recording."
    )
    _add_arguments_check(parser_check)

    # parser for the command napari
    parser_napari = subparsers.add_parser(
        name="napari", description="Render localization data in napari."
    )
    _add_arguments_napari(parser_napari)

    # parser for the command show_versions
    parser_show_versions = subparsers.add_parser(
        name="show_versions",
        description="Show system information and dependency versions.",
    )
    _add_arguments_show_versions(parser_show_versions)

    # parser for the command test
    parser_test = subparsers.add_parser(name="test", description="Run test suite.")
    _add_arguments_test(parser_test)

    # Parse
    returned_args = parser.parse_args(args)

    if returned_args.info:
        show_versions()

    elif returned_args.command:
        if returned_args.command == "draw_roi_mpl":
            from .scripts.script_draw_roi import sc_draw_roi_mpl

            sc_draw_roi_mpl(
                returned_args.file,
                returned_args.type,
                returned_args.roi_file_indicator,
                returned_args.region_type,
            )

        elif returned_args.command == "rois":
            from .scripts.script_rois import sc_draw_roi_napari

            sc_draw_roi_napari(
                file_path=returned_args.file,
                file_type=returned_args.type,
                roi_file_indicator=returned_args.roi_file_indicator,
                bin_size=returned_args.bin_size,
                rescale=returned_args.rescale,
            )

        elif returned_args.command == "check":
            from .scripts.script_check import sc_check

            sc_check(
                pixel_size=returned_args.pixel_size,
                file_images=returned_args.file_images,
                file_locdata=returned_args.file_locdata,
                file_type=returned_args.file_type,
                transpose=True,
                kwargs_image={},
                kwargs_points={},
            )

        elif returned_args.command == "napari":
            from .scripts.script_napari import sc_napari

            sc_napari(
                file_path=returned_args.file,
                file_type=returned_args.type,
                bin_size=returned_args.bin_size,
                rescale=returned_args.rescale,
            )

        elif returned_args.command == "show_versions":
            from .scripts.script_show_versions import sc_show_versions  # type: ignore[attr-defined]  # noqa

            sc_show_versions(
                verbose=returned_args.verbose,
                extra_dependencies=returned_args.extra,
                other_dependencies=returned_args.other,
            )

        elif returned_args.command == "test":
            from .scripts.script_test import sc_test  # type: ignore[attr-defined]

            sc_test(args=returned_args.pytest_args)

    else:
        print(
            "This is the command line entry point for locan. Get more information with 'locan -h'."
        )

    return None


if __name__ == "__main__":
    sys.exit(main())  # type: ignore[func-returns-value]



"""

Test the locan package.

This module includes unit tests for all modules within the locan package.
The tests are organized following the subpackage structure of locan.

"""

from __future__ import annotations

from typing import TYPE_CHECKING

from locan import ROOT_DIR

if TYPE_CHECKING:
    from pytest import ExitCode

__all__: list[str] = ["test"]


def test(args: str | list[str] | None = None) -> int | ExitCode:
    """
    Running tests with pytest.

    Parameters
    ----------
    args
        Parameters passed to :func:`pytest.main`
    """
    try:
        import pytest
    except ImportError as exc:
        raise ImportError("Need pytest to run tests.") from exc

    extra_args = ["-m not gui and not visual"]

    if args is None:
        pass  # extra_args = []
    elif isinstance(args, list):
        extra_args.extend(args)  # extra_args = args
    else:
        extra_args.append(args)  # extra_args = [args]

    test_directory = ROOT_DIR.joinpath("tests").as_posix()
    extra_args.append(test_directory)
    print(f'running: pytest {" ".join(extra_args)}')
    return pytest.main(extra_args)



import pytest

from locan.datasets import load_npc, load_tubulin

httpx = pytest.importorskip("httpx", reason="requires httpx for html download")


def test_load_npc():
    locdata = load_npc()
    assert locdata.meta.element_count == 2285189


def test_load_tubulin():
    locdata = load_tubulin()
    assert locdata.meta.element_count == 1506568



"""

Deal with metadata in LocData objects.

Functions to modify metadata in LocData objects.

"""

from __future__ import annotations

import importlib
import logging
import os
from typing import TYPE_CHECKING, Any, BinaryIO

try:
    import tomllib  # type: ignore[import]
except ModuleNotFoundError:
    import tomli as tomllib  # type: ignore  # for sys.version_info < (3, 11):

from google.protobuf import json_format, text_format
from google.protobuf.message import Message

from locan.data import metadata_pb2

if TYPE_CHECKING:
    from locan.data.locdata import LocData

__all__: list[str] = [
    "metadata_to_formatted_string",
    "metadata_from_toml_string",
    "load_metadata_from_toml",
    "message_scheme",
    "merge_metadata",
]

logger = logging.getLogger(__name__)


def _modify_meta(
    locdata: LocData,
    new_locdata: LocData,
    function_name: str | None = None,
    parameter: dict[str, Any] | None = None,
    meta: metadata_pb2.Metadata | None = None,
) -> metadata_pb2.Metadata:
    """
    Update metadata in Locdata after modification of locdata.

    Parameters
    ----------
    locdata
        original locdata before modification
    new_locdata
        new locdata after modification
    function_name
        Name of function that was applied for modification.
    parameter
        Parameter for function that was applied for modification.
    meta
        Metadata about the current dataset and its history.

    Returns
    -------
    locan.data.metadata_pb2.Metadata
        Metadata about the current dataset and its history.
    """
    meta_ = metadata_pb2.Metadata()
    meta_.CopyFrom(locdata.meta)
    # try:
    #     meta_.ClearField("identifier")
    # except ValueError:
    #     pass
    #
    # try:
    #     meta_.ClearField("element_count")
    # except ValueError:
    #     pass
    #
    # try:
    #     meta_.ClearField("frame_count")
    # except ValueError:
    #     pass

    meta_.identifier = new_locdata.meta.identifier
    meta_.element_count = new_locdata.meta.element_count
    meta_.frame_count = new_locdata.meta.frame_count
    meta_.modification_time.CopyFrom(new_locdata.meta.creation_time)

    meta_.state = metadata_pb2.MODIFIED
    meta_.ancestor_identifiers.append(locdata.meta.identifier)
    meta_.history.add(name=function_name, parameter=str(parameter))

    if meta is None:
        pass
    elif isinstance(meta, dict):
        for key, value in meta.items():
            setattr(meta_, key, value)
    else:
        meta_.MergeFrom(meta)

    return meta_


def _dict_to_protobuf(
    dictionary: dict[str, Any], message: Message, inplace: bool = False
) -> Message | None:
    """
    Parse dictionary with message attributes and their values in message.
    """
    if inplace is False:
        message_ = message.__class__()
        message_.CopyFrom(message)
        message = message_

    for key, value in dictionary.items():
        try:
            attr_ = getattr(message, key)
        except AttributeError as e:
            logger.warning(f"AttributeError while parsing: {e}")
            break

        if isinstance(value, dict):
            _dict_to_protobuf(dictionary=value, message=attr_, inplace=True)
        elif isinstance(value, list):
            try:
                attr_.extend(value)
            except TypeError:
                for element in value:
                    submessage = attr_.add()
                    _dict_to_protobuf(
                        dictionary=element, message=submessage, inplace=True
                    )
        else:
            try:
                setattr(message, key, value)
            except AttributeError:
                if attr_.DESCRIPTOR.name == "Timestamp":
                    attr_.FromJsonString(value)
                elif attr_.DESCRIPTOR.name == "Duration":
                    attr_.FromNanoseconds(value)

    if inplace:
        return None
    else:
        return message


def metadata_to_formatted_string(message: Message, **kwargs: Any) -> str:
    """
    Get formatted string from Locdata.metadata.

    Parameters
    ----------
    message
        Protobuf message like locan.data.metadata_pb2.Metadata
    kwargs
        Other kwargs that are passed to
        :func:`google.protobuf.text_format.MessageToString`.

    Returns
    -------
    str
        Formatted metadata string.
    """

    def message_formatter(
        message: Message, indent: int, as_one_line: bool
    ) -> str | None:
        if message.DESCRIPTOR.name in ["Timestamp", "Duration"]:
            return_value: str | None = message.ToJsonString()  # type: ignore[attr-defined]
        else:
            return_value = None
        return return_value

    return text_format.MessageToString(
        message, message_formatter=message_formatter, **kwargs
    )


def _toml_dict_to_protobuf(toml_dict: dict[str, Any]) -> dict[str, Message]:
    """
    Turn toml dict into protobuf messages.

    Parameters
    ----------
    toml_dict
        Dict from TOML string with metadata.

    Returns
    -------
    dict[str, google.protobuf.message.Message]
        Message instances with name as declared in toml file.
    """
    # instantiate messages
    instances: dict[str, Message] = {}
    for message in toml_dict.pop("messages"):
        module = importlib.import_module(message["module"])
        class_ = getattr(module, message["class_name"])
        instances[message["name"]] = class_()

    # parse values
    for message_name, dictionary in toml_dict.items():
        _dict_to_protobuf(
            dictionary=dictionary, message=instances[message_name], inplace=True
        )

    return instances


def metadata_from_toml_string(toml_string: str | None) -> dict[str, Message] | None:
    """
    Turn toml string into protobuf message instances.

    Note
    -----
    Parses Timestamp elements from string '2022-05-14T06:58:00Z'.
    Parses Duration elements from int in nanoseconds.

    Parameters
    ----------
    toml_string
        TOML string with metadata.

    Returns
    -------
    dict[str, google.protobuf.message.Message] | None
        Message instances with name as declared in toml file.
    """
    if toml_string is None:
        return None

    toml_dict = tomllib.loads(toml_string)
    return _toml_dict_to_protobuf(toml_dict)


def load_metadata_from_toml(
    path_or_file_like: str | bytes | os.PathLike[Any] | BinaryIO | None,
) -> dict[str, Message] | None:
    """
    Turn toml file into protobuf message instances.

    Note
    -----
    Parses Timestamp elements from string '2022-05-14T06:58:00Z'.
    Parses Duration elements from int in nanoseconds.

    Parameters
    ----------
    path_or_file_like
        File path or file-like for a TOML file.

    Returns
    -------
    dict[str, google.protobuf.message.Message] | None
        Message instances with name as declared in toml file.
    """
    if path_or_file_like is None:
        return None

    try:
        toml_dict = tomllib.load(path_or_file_like)  # type: ignore[arg-type]
    except AttributeError:
        with open(path_or_file_like, "rb") as file:  # type: ignore[arg-type]
            toml_dict = tomllib.load(file)

    return _toml_dict_to_protobuf(toml_dict)


def message_scheme(message: Message) -> dict[str, Any]:
    """
    Provide message scheme with defaults including nested messages.

    Parameters
    ----------
    message
        Protobuf message

    Returns
    -------
    dict[str, Any]
        A nested dictionary with all message fields including default values.
    """

    message_dict = json_format.MessageToDict(
        message,
        always_print_fields_with_no_presence=True,
        preserving_proto_field_name=True,
    )

    for descriptor in message.DESCRIPTOR.fields:
        if descriptor.type == descriptor.TYPE_MESSAGE:
            attr_ = getattr(message, descriptor.name)

            if descriptor.label != descriptor.LABEL_REPEATED:
                message_dict[descriptor.name] = message_scheme(attr_)

            elif (
                descriptor.label == descriptor.LABEL_REPEATED
                and "ScalarMap" not in type(attr_).__name__
                and "MessageMapContainer" not in type(attr_).__name__
            ):
                attr_ = attr_.add()
                message_dict[descriptor.name] = message_scheme(attr_)

    return message_dict


def merge_metadata(
    metadata: metadata_pb2.Metadata | None = None,
    other_metadata: (
        metadata_pb2.Metadata
        | None
        | dict[str, Any]
        | str
        | bytes
        | os.PathLike[Any]
        | BinaryIO
    ) = None,
) -> metadata_pb2.Metadata:
    """
    Merge `other_metadata` into Locdata.meta.

    Parameters
    ----------
    metadata
        Original LocData metadata before modification
    other_metadata
        Metadata to be merged.

    Returns
    -------
    locan.data.metadata_pb2.Metadata
        Merged metadata
    """
    if metadata is None:
        new_metadata = metadata_pb2.Metadata()
    else:
        new_metadata = metadata

    if other_metadata is None:
        pass
    elif isinstance(other_metadata, (str, bytes, os.PathLike, BinaryIO)):
        meta_ = load_metadata_from_toml(other_metadata)
        if meta_ is not None:
            new_metadata.MergeFrom(meta_["metadata"])  # type: ignore[arg-type]
    elif isinstance(other_metadata, dict):
        for key, value in other_metadata.items():
            setattr(new_metadata, key, value)
    else:
        new_metadata.MergeFrom(other_metadata)

    return new_metadata



import pytest

import locan.constants
import locan.locan_io.locdata.io_locdata as io
from locan import LocData
from locan.analysis import AccumulationClusterCheck
from locan.analysis.accumulation_analysis import (
    _accumulation_cluster_check,
    _accumulation_cluster_check_for_single_dataset,
)


@pytest.fixture()
def locdata_blobs():
    return io.load_txt_file(path=locan.ROOT_DIR / "tests/test_data/five_blobs.txt")


@pytest.fixture()
def locdata_blobs_3D():
    return io.load_txt_file(path=locan.ROOT_DIR / "tests/test_data/five_blobs_3D.txt")


def test___accumulation_cluster_check_for_single_dataset(locdata_blobs):
    res = _accumulation_cluster_check_for_single_dataset(
        locdata_blobs,
        region_measure=locdata_blobs.bounding_box.region_measure,
        hull="bb",
    )
    assert len(res) == 3
    res = _accumulation_cluster_check_for_single_dataset(
        locdata_blobs,
        region_measure=locdata_blobs.bounding_box.region_measure,
        hull="ch",
    )
    assert len(res) == 3


def test__accumulation_cluster_check(locdata_blobs):
    results = _accumulation_cluster_check(locdata_blobs)
    assert len(results.columns) == 5

    results = _accumulation_cluster_check(
        locdata_blobs, n_loc=3, hull="bb", divide="sequential"
    )
    assert len(results.columns) == 5

    results = _accumulation_cluster_check(
        locdata_blobs, n_loc=3, hull="ch", divide="sequential"
    )
    assert len(results.columns) == 5

    results = _accumulation_cluster_check(
        locdata_blobs, region_measure=100, n_loc=3, divide="sequential"
    )
    assert len(results.columns) == 5

    results = _accumulation_cluster_check(
        locdata_blobs, n_loc=[20, 30], divide="sequential"
    )
    assert len(results.columns) == 5


def test_AccumulationClusterCheck_empty(caplog):
    acc = AccumulationClusterCheck().compute(LocData())
    acc.plot()
    assert caplog.record_tuples == [
        ("locan.analysis.accumulation_analysis", 30, "Locdata is empty.")
    ]


def test_AccumulationClusterCheck(locdata_blobs):
    acc = AccumulationClusterCheck(
        algo_parameter=dict(min_cluster_size=4, allow_single_cluster=False), n_loc=5
    ).compute(locdata_blobs)
    assert len(acc.results.columns) == 5
    acc = AccumulationClusterCheck(
        algo_parameter=dict(min_cluster_size=4, allow_single_cluster=False),
        n_loc=5,
        n_extrapolate=2,
    ).compute(locdata_blobs)
    assert len(acc.results.columns) == 5

    acc.plot()



"""

User interfaces.

This module provides functions and classes for using graphical user interfaces (GUI).

Functions provide a GUI based on QT if the QT backend and appropriate python bindings
are available. Supported bindings are found in the enum class `locan.QtBindings`.

The configuration variable `locan.QT_BINDING` declares which python binding to use
if several are installed.
If an environment variable `QT_API` is defined it takes precedence over
`locan.QT_BINDING`.

If neither `locan.QT_BINDING` nor `QT_API` is defined, qtpy will choose the binding.

Submodules:
-----------

.. autosummary::
   :toctree: ./

   io

"""

from __future__ import annotations

from locan.gui import io

from .io import *

__all__: list[str] = []
__all__.extend(io.__all__)



"""

File input/output for localization data in DECODE files.

"""

from __future__ import annotations

import logging
import os
from collections.abc import Mapping
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from _typeshed import SupportsRead

import pandas as pd

import locan.constants
from locan.data import metadata_pb2
from locan.data.locdata import LocData
from locan.dependencies import HAS_DEPENDENCY, needs_package
from locan.locan_io.locdata.utilities import (
    convert_property_names,
    convert_property_types,
)

if HAS_DEPENDENCY["h5py"]:
    import h5py


__all__: list[str] = ["load_decode_header", "load_decode_file"]

logger = logging.getLogger(__name__)


@needs_package("h5py")
def _read_decode_header(
    file: Mapping[str, Any]
) -> tuple[list[str], dict[str, Any], dict[str, Any]]:
    """
    Read header from a DECODE single-molecule localization file and identify
    column names.

    Parameters
    ----------
    file
        HDF5 file object.

    Returns
    -------
    tuple[list[str], dict[str, Any], dict[str, Any]]
        Tuple with identifiers, meta and decode sections.
        Identifiers are list of valid dataset property keys as derived from
        the DECODE identifiers.
    """
    meta_data = dict(file["meta"].attrs)
    meta_decode = dict(file["decode"].attrs)

    # list identifiers
    identifiers = list(file["data"].keys())

    column_keys = []
    for i in identifiers:
        if i == "xyz":
            column_keys.extend(["position_x", "position_y", "position_z"])
        elif i == "xyz_cr":
            column_keys.extend(["x_cr", "y_cr", "z_cr"])
        elif i == "xyz_sig":
            column_keys.extend(["x_sig", "y_sig", "z_sig"])
        else:
            column_keys.append(i)

    column_keys = convert_property_names(
        properties=column_keys, property_mapping=locan.constants.DECODE_KEYS
    )

    return column_keys, meta_data, meta_decode


@needs_package("h5py")
def load_decode_header(
    path: str | os.PathLike[Any] | SupportsRead[Any],
) -> tuple[list[str], dict[str, Any], dict[str, Any]]:
    """
    Load header from a DECODE single-molecule localization file and identify
    column names.

    The hdf5 file should contain the following keys:
    <KeysViewHDF5 ['data', 'decode', 'meta']>

    Parameters
    ----------
    path : str | os.PathLike | SupportsRead
        File path or file-like object for a file to load.

    Returns
    -------
    tuple[list[str], dict[str, Any], dict[str, Any]]
        Tuple with identifiers, meta and decode sections.
        Identifiers are list of valid dataset property keys as derived from
        the DECODE identifiers.
    """
    with h5py.File(path, "r") as file:
        return _read_decode_header(file)


@needs_package("h5py")
def load_decode_file(
    path: str | os.PathLike[Any] | SupportsRead[Any],
    nrows: int | None = None,
    convert: bool = True,
) -> LocData:
    """
    Load data from a DECODE single-molecule localization file.

    Parameters
    ----------
    path
        File path or file-like object for a file to load.
    nrows
        The number of localizations to load from file.
        None means that all available rows are loaded.
    convert
        If True convert types by applying type specifications in
        locan.constants.PROPERTY_KEYS.

    Returns
    -------
    LocData
        A new instance of LocData with all localizations.
    """
    with h5py.File(path, "r") as file:
        columns, meta, _ = _read_decode_header(file)

        if file["data"]["xyz"].shape == (0, 3):  # empty file
            logger.warning("File does not contain any data.")
            locdata = LocData()

        else:  # file not empty
            data = {}
            for key, value in file["data"].items():
                if value.shape is not None:
                    if key == "xyz":
                        for i, property_ in enumerate(
                            ["position_x", "position_y", "position_z"]
                        ):
                            data[property_] = value[:nrows, i]
                    elif key == "xyz_cr":
                        for i, property_ in enumerate(["x_cr", "y_cr", "z_cr"]):
                            data[property_] = value[:nrows, i]
                    elif key == "xyz_sig":
                        for i, property_ in enumerate(["x_sig", "y_sig", "z_sig"]):
                            data[property_] = value[:nrows, i]
                    elif key in locan.constants.DECODE_KEYS:
                        data[locan.constants.DECODE_KEYS[key]] = value[:nrows]
                    else:
                        data[key] = value[:nrows]

            dataframe = pd.DataFrame(data)

            if convert:
                dataframe = convert_property_types(
                    dataframe, types=locan.constants.PROPERTY_KEYS
                )

            locdata = LocData.from_dataframe(dataframe=dataframe)

    locdata.meta.source = metadata_pb2.EXPERIMENT
    locdata.meta.state = metadata_pb2.RAW
    locdata.meta.file.type = metadata_pb2.DECODE
    locdata.meta.file.path = str(path)

    for property_ in sorted(
        list(set(columns).intersection({"position_x", "position_y", "position_z"}))
    ):
        locdata.meta.localization_properties.add(
            name=property_, unit=meta["xy_unit"], type="float"
        )

    del locdata.meta.history[:]
    locdata.meta.history.add(
        name="load_decode_file", parameter=f"path={path}, nrows={nrows}"
    )

    return locdata



"""

Statistics related tools.

"""

from __future__ import annotations

import logging
from typing import NamedTuple

import numpy as np
import numpy.typing as npt

__all__: list[str] = [
    "weighted_mean_variance",
    "ratio_fwhm_to_sigma",
    "biased_variance",
]

logger = logging.getLogger(__name__)


class WeightedMeanVariance(NamedTuple):
    weighted_mean: float | npt.NDArray[np.float64]
    weighted_mean_variance: float | npt.NDArray[np.float64]


def weighted_mean_variance(
    values: npt.ArrayLike, weights: npt.ArrayLike | None
) -> WeightedMeanVariance:
    """
    Compute weighted mean (average)
    and the corresponding weighted mean variance [1]_.

    Parameters
    ----------
    values
        Values from which to compute the weighted average.
    weights
        Weights to use for weighted average.

    Returns
    -------
    WeightedMeanVariance
        weighted_mean, weighted_mean_variance

    References
    ----------
    .. [1] Donald F. Gatz, Luther Smith,
       The standard error of a weighted mean concentration —
       I. Bootstrapping vs other methods.
       Atmospheric Environment 29(11), 1995: 1185-1193,
       https://doi.org/10.1016/1352-2310(94)00210-C.
    """
    values = np.asarray(values)
    if values.ndim > 1:
        raise ValueError("values must be 1-dimensional.")
    if weights is None:
        weights = np.ones(shape=values.shape)
    else:
        weights = np.asarray(weights)
    if values.shape != weights.shape:
        raise TypeError("Shape of values and weights must be the same.")

    n_weights = len(weights)
    if n_weights == 1:
        weighted_average = np.mean(a=values)
        weighted_mean_variance_ = 0
    else:
        weighted_average = np.average(a=values, weights=weights)
        weights_mean = np.mean(a=weights)
        weighted_mean_variance_ = (
            n_weights
            / (n_weights - 1)
            / np.sum(weights) ** 2
            * (
                sum((weights * values - weights_mean * weighted_average) ** 2)
                - 2
                * weighted_average
                * sum(
                    (weights - weights_mean)
                    * (weights * values - weights_mean * weighted_average)
                )
                + weighted_average**2 * sum((weights - weights_mean) ** 2)
            )
        )

        # Check for negative values that should not occur
        # but did result with some floating values
        if (
            weighted_mean_variance_ != 0
            and weighted_mean_variance_[weighted_mean_variance_ < 0].any()  # type: ignore
        ):
            if np.ndim(weighted_mean_variance_) == 0:
                weighted_mean_variance_ = 0
            else:
                weighted_mean_variance_[weighted_mean_variance_ < 0] = 0  # type: ignore
            logger.warning(
                "Negative values for weighted_mean_variance occurred and were set to "
                "zero."
            )

    return WeightedMeanVariance(
        weighted_mean=weighted_average,
        weighted_mean_variance=weighted_mean_variance_,
    )


def ratio_fwhm_to_sigma() -> float:
    """
    The numeric value of the ratio between full-width-half-max (fwhm) width
    and standard deviation (sigma) of a normal distribution.

    Returns
    -------
    float
    """
    return_value = 2 * np.sqrt(2 * np.log(2))
    return float(return_value)


def biased_variance(
    variance: npt.ArrayLike, n_samples: npt.ArrayLike
) -> npt.NDArray[np.float64]:
    """
    The sample variance is biased if not corrected by Bessel's correction.
    This function yields the biased variance by applying the inverse
    correction.

    .. math::

        E(variance_{biased}) = variance * (1 - 1 / localization_counts).

    Parameters
    ----------
    variance
        An unbiased variance.
    n_samples
        Number of samples from which the biased sample variance would be
        computed.

    Returns
    -------
    npt.NDArray[np.float64]
    """
    variance = np.asarray(variance)
    n_samples = np.asarray(n_samples)
    return_value: npt.NDArray[np.float64] = variance * (1 - 1 / n_samples)
    return return_value



import boost_histogram as bh
import matplotlib.pyplot as plt
import pandas as pd
import pytest

from locan import LocData
from locan.analysis import PositionVarianceExpectation
from locan.analysis.position_variance_expectation import (
    PositionVarianceExpectationResults,
    _property_variances,
)


def test__property_variances(locdata_2d):
    collection = LocData.from_collection(locdatas=[locdata_2d, locdata_2d])
    result = _property_variances(collection=collection, loc_property="position_x")
    assert all(result["localization_count"] == [6, 6])
    assert all(
        result["position_x_var"] == [pytest.approx(2.666667), pytest.approx(2.666667)]
    )


class TestPositionVarianceExpectation:
    def test_init(self, locdata_blobs_2d):
        gpe = PositionVarianceExpectation(meta={"comment": "this is an example"})
        assert (
            str(gpe) == "PositionVarianceExpectation("
            "loc_property=position_x, expectation=None, biased=True)"
        )
        assert gpe.results is None
        assert gpe.meta.comment == "this is an example"

        gpe = PositionVarianceExpectation(
            meta={"comment": "this is an example"},
            loc_property="position_y",
        )
        assert (
            str(gpe) == "PositionVarianceExpectation("
            "loc_property=position_y, expectation=None, biased=True)"
        )
        assert gpe.results is None
        assert gpe.meta.comment == "this is an example"

    def test_empty_locdata(self, caplog):
        gpe = PositionVarianceExpectation().compute(LocData())
        assert gpe.results is None
        gpe.plot()
        gpe.hist()
        # plt.show()

        assert caplog.record_tuples == [
            ("locan.analysis.position_variance_expectation", 30, "Locdata is empty."),
        ]

        plt.close("all")

    def test_compute(self, locdata_2d):
        collection = LocData.from_collection(locdatas=[locdata_2d, locdata_2d])
        pve = PositionVarianceExpectation().compute(locdata=collection)
        assert pve.results.values["position_x_var"].tolist() == pytest.approx(
            [2.2222222222222228, 2.2222222222222228], rel=0.1
        )
        # print(pve.results.values)
        # print(pve.results.grouped)
        assert isinstance(pve.results, PositionVarianceExpectationResults)
        assert pve.results.values.index.tolist() == collection.data.index.tolist()
        assert pve.results.values.columns.tolist() == [
            "localization_count",
            "position_x_var",
            "expectation",
            "value_to_expectation_ratio",
        ]
        assert pve.results.grouped.index.tolist() == [6]
        assert pve.results.grouped.columns.tolist() == [
            "position_x_var_mean",
            "position_x_var_std",
            "expectation",
        ]
        assert pve.results.grouped.expectation.isna().all()

        pve.plot()
        # plt.show()
        pve.hist()
        # plt.show()

        pve = PositionVarianceExpectation(biased=False).compute(locdata=collection)
        assert pve.results.values["position_x_var"].tolist() == pytest.approx(
            [2.67, 2.67], rel=0.1
        )

        plt.close("all")

    def test_compute_with_expectation(self, locdata_2d):
        # biased
        collection = LocData.from_collection(locdatas=[locdata_2d, locdata_2d])
        expected = 2
        gpe = PositionVarianceExpectation(
            expectation=expected,
        ).compute(locdata=collection)
        assert gpe.results.values.index.tolist() == collection.data.index.tolist()
        assert gpe.results.values.columns.tolist() == [
            "localization_count",
            "position_x_var",
            "expectation",
            "value_to_expectation_ratio",
        ]
        assert gpe.results.grouped.index.tolist() == [6]
        assert gpe.results.grouped.columns.tolist() == [
            "position_x_var_mean",
            "position_x_var_std",
            "expectation",
        ]
        assert gpe.results.grouped.expectation.tolist() == [
            pytest.approx(1.666667, rel=1e-5)
        ]
        assert all(
            gpe.results.values.value_to_expectation_ratio
            == [pytest.approx(1.33333, rel=1e-5), pytest.approx(1.33333, rel=1e-5)]
        )

        # unbiased
        collection = LocData.from_collection(locdatas=[locdata_2d, locdata_2d])
        expected = 2
        gpe = PositionVarianceExpectation(expectation=expected, biased=False).compute(
            locdata=collection
        )
        assert gpe.results.values.index.tolist() == collection.data.index.tolist()
        assert gpe.results.values.columns.tolist() == [
            "localization_count",
            "position_x_var",
            "expectation",
            "value_to_expectation_ratio",
        ]
        assert gpe.results.grouped.index.tolist() == [6]
        assert gpe.results.grouped.columns.tolist() == [
            "position_x_var_mean",
            "position_x_var_std",
            "expectation",
        ]
        assert all(gpe.results.grouped.expectation == 2)
        assert all(
            gpe.results.values.value_to_expectation_ratio
            == gpe.results.values.position_x_var / expected
        )

        expected = pd.Series(data=[2], index=[6])
        gpe = PositionVarianceExpectation(expectation=expected, biased=False).compute(
            locdata=collection
        )
        assert gpe.results.values.index.tolist() == collection.data.index.tolist()
        assert gpe.results.values.columns.tolist() == [
            "localization_count",
            "position_x_var",
            "expectation",
            "value_to_expectation_ratio",
        ]
        assert gpe.results.grouped.index.tolist() == [6]
        assert gpe.results.grouped.columns.tolist() == [
            "position_x_var_mean",
            "position_x_var_std",
            "expectation",
        ]
        assert all(gpe.results.grouped.expectation == 2)
        assert all(
            gpe.results.values.value_to_expectation_ratio
            == gpe.results.values.position_x_var / 2
        )

        expected = {6: 2}
        gpe = PositionVarianceExpectation(expectation=expected, biased=False).compute(
            locdata=collection
        )
        assert gpe.results.values.index.tolist() == collection.data.index.tolist()
        assert gpe.results.grouped.index.tolist() == [6]
        assert all(gpe.results.grouped.expectation == 2)
        assert all(
            gpe.results.values.value_to_expectation_ratio
            == gpe.results.values.position_x_var / 2
        )

        gpe.plot()
        # plt.show()
        gpe.hist()
        # plt.show()

        plt.close("all")


@pytest.mark.visual
class TestPositionVarianceExpectationVisual:
    def test_compute(self, locdata_blobs_2d):
        grouped = locdata_blobs_2d.data.groupby("cluster_label")
        collection = LocData.from_collection(
            locdatas=[
                LocData.from_dataframe(dat.iloc[: (10 - i)])
                for i, (name, dat) in enumerate(grouped)
            ]
        )
        pve = PositionVarianceExpectation(
            expectation=400,
        ).compute(locdata=collection)
        assert isinstance(pve.results, PositionVarianceExpectationResults)
        print(pve.results.values.columns)
        print(pve.results.grouped)

        pve.plot()
        plt.show()

        pve.hist()
        plt.show()

        pve.hist(bin_size=(1, 1000), bin_range=((3, 100), (1, 10_000)), log=False)
        plt.show()
        pve.hist(bin_size=(1, 1000), bin_range=((3, 100), (1, 10_000)))
        plt.show()

        axes = bh.axis.AxesTuple(
            (
                bh.axis.Regular(20, 3, 503, transform=bh.axis.transform.log),
                bh.axis.Regular(50, 1, 15_000, transform=bh.axis.transform.log),
            )
        )
        pve.hist(bins=axes)
        plt.show()

        plt.close("all")
